# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from flask import request, jsonify

import api
from api.core import API, secure
from api.security import checkPermissions

from services import Service, ServiceHub
from services.ldap import LdapService
from tools import mconf
from tools.permissions import SystemAdminPermission, SystemAdminROPermission


@API.route(api.BaseRoute+"/system/mconf/ldap", methods=["GET", "DELETE"])
@secure(requireDB=True, authLevel="user")
def getLdapConfig():
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    if request.method == "DELETE":
        mconf.dumpLdap({})
        ServiceHub.load("ldap", force_reload=True)
        return jsonify(message="LDAP deactivated")
    return jsonify(ldapAvailable=Service.available("ldap"), data=mconf.LDAP)


@API.route(api.BaseRoute+"/system/mconf/ldap", methods=["PUT"])
@secure(requireDB=True, authLevel="user")
def setLdapConfig():
    checkPermissions(SystemAdminPermission())
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing configuration"), 400
    error = LdapService.testConfig(data)
    forced = False
    if error:
        if request.args.get("force") == "true":
            forced = True
        else:
            return jsonify(message=error), 400
    error = mconf.dumpLdap(data)
    if error:
        return jsonify(message="Configuration updated, but save to disk failed: "+error), 500
    ServiceHub.load("ldap", force_reload=True)
    return jsonify(message="Force updated LDAP configuration" if forced else "LDAP configuration updated")


@API.route(api.BaseRoute+"/system/mconf/authmgr", methods=["GET", "DELETE"])
@secure(requireDB=True, authLevel="user")
def getAuthmgrConfig():
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    if request.method == "DELETE":
        mconf.dumpAuthmgr({})
        return jsonify(message="authmgr configuration set to default")
    return jsonify(data=mconf.AUTHMGR)


@API.route(api.BaseRoute+"/system/mconf/authmgr", methods=["PUT"])
@secure(requireDB=True, authLevel="user")
def setAuthmgrConfig():
    checkPermissions(SystemAdminPermission())
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing configuration"), 400
    error = mconf.dumpAuthmgr(data)
    if error:
        return jsonify(message="Configuration updated, but save to disk failed: "+error), 500
    return jsonify(message="authmgr configuration updated")



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

import argcomplete
import logging
from argparse import ArgumentParser, _SubParsersAction


class CliError(BaseException):
    pass


class InvalidUseError(Exception):
    pass


class Cli:
    class Formatter(logging.Formatter):
        levelstyles = {"DEBUG": {"attrs": ["dark"]},
                       "WARNING": {"color": "yellow"},
                       "ERROR": {"color": "red"},
                       "CRITICAL": {"color": "red", "attrs": ["bold"]}}

        def __init__(self, format, col):
            super().__init__(format)
            self.__col = col

        def format(self, record):
            record.msg = self.__col(record.msg, **self.levelstyles.get(record.levelname, {}))
            return logging.Formatter.format(self, record)

    funcs = []

    def __init__(self, mode="standalone", **kwargs):
        """Create cli object

        Parameters
        ----------
        mode : str, optional
            Cli execution mode. Applies some global adjustments if set to "standalone".
            No other setting currently has an effect.
            The default is "standalone".

        Keyword Arguments
        -----------------

        color : bool, optional
            Force color mode.
            Default is to auto-detect if color is available.
        completer : object, optional
            Object providing a `rl_complete(prefix, state)` function used for command completion.
            Default argcomplete.Completer()
        fs : dict, optional
            If set, use as base for emulated filesystem.
            May contain name -> content mapping to emulate readable files.
            `closeFiles` will automatically remove readable files and store written files along with their write mode as dicts.
            Default is None (write directly to disk)
        host : str, optional
            Host string to display next to the command prompt. Default is None.
        stdin : file-like, optional
            If set, use this file as stdin.
            Defaults to `sys.stdin`
        stdout : file-like, optional
            If set, use this file as stdout
            Defaults to `sys.stdout`

        **kwargs : any
            Further keyword arguments
        """
        self._createParser()
        if mode == "standalone":
            argcomplete.autocomplete(self.parser)
        import sys
        self.mode = mode
        self.stdout = kwargs.get("stdout", sys.stdout)
        self.stdin = kwargs.get("stdin", sys.stdin)
        self.host = kwargs.get("host")
        self.completer = kwargs.get("completer")
        self.__completing = True
        self.colored = self.detectcolor(self.stdout, kwargs.get("color"))
        self.col = self.colorfunc(self.colored)
        self.fs = None
        if "fs" in kwargs and isinstance(kwargs["fs"], dict):
            self.fs = kwargs["fs"]
            import io
            for file, content in self.fs.items():
                self.fs[file] = dict(mode="r",
                                     stream=(io.BytesIO if isinstance(content, bytes) else io.StringIO)(content))
        if mode == "standalone":
            self.initLogging(self.stdout, color=self.colored)

    def execute(self, args=None, secure=True):
        """Execute commands as specified by args.

        If no args are passed, sys.args is used instead.

        Parameters
        ----------
        args : list of strings, optional
            Command line arguments to execute. The default is None.
        """
        self.__completing = False
        dispatch = self.parser.parse_args(args)
        self.__completing = True
        dispatch._cli = self
        if hasattr(dispatch, "_handle"):
            try:
                return dispatch._handle(dispatch) or 0
            except CliError as err:
                self.print(self.col(err.args[0], "red"))
                return 1
            except InvalidUseError:
                self.print(self.col("Missing command. Use --help for more info.", "red"))
                return 3
            except BaseException as err:
                if not secure:
                    raise
                self.print(self.col("An error occurred: {}."
                                    .format(repr(err)), "red"))
                return -1
        else:
            self.parser.print_help()
            return 2

    @classmethod
    def register(cls, name, handler, setup, **kwargs) -> ArgumentParser:
        """Register a new sub-command.

        The parsed arguments are passed to the handler function.

        Parameters
        ----------
        name : str
            Name of the sub-command
        handler : Callable
            Function that executes the sub-command.
        kwargs : any
            Arguments passed to the add_parser function

        Returns
        -------
        ArgumentParser
            Sub-parser that can be customized with sub-command specific arguments
        """
        cls.funcs.append((name, handler, setup, kwargs))

    @classmethod
    def command(cls, name, parserSetup=lambda subp: None, **kwargs):
        """Decorator for sub-command handlers.

        Can be used instead of calling register().

        Parameters
        ----------
        name : str
            Name of the subcommand
        parserSetup : Callable, optional
            Function that sets up the sub-command parser. By default no further initialization is done.
        kwargs : any
            Arguments passed to the add_parser function
        """
        def inner(func):
            cls.register(name, func, parserSetup, **kwargs)
            return func
        return inner

    @staticmethod
    def colorfunc(col):
        if col:
            import termcolor
            return termcolor.colored
        else:
            return lambda text, *args, **kwargs: text

    @staticmethod
    def detectcolor(stdout, preset=None):
        try:
            import termcolor
            if preset is not None:
                return preset
            return stdout.isatty()
        except Exception:
            return False

    @staticmethod
    def initLogging(stdout=None, color=None):
        """Initialize stand-alone mode."""
        if stdout is None:
            import sys
            stdout = sys.stdout
        color = Cli.detectcolor(stdout, color)
        fmt = "(%(name)s) %(message)s"
        if not color:
            fmt = "[%(levelname)s] "+fmt
        handler = logging.StreamHandler()
        handler.setFormatter(Cli.Formatter(fmt, Cli.colorfunc(color)))
        logging.root.handlers = [handler]
        from tools.config import initLoggers
        try:
            initLoggers()
        except Exception as err:
            logging.getLogger("config").error("Failed to initialize loggers: "+" - ".join(str(arg) for arg in err.args))

    def _createParser(self):
        """Create parser from registered functions."""
        def redirect(parser):
            def perr(msg):
                if not self.__completing:
                    parser.print_usage(self.stdout)
                    self.print(msg)
                raise SystemExit(1)

            print_help = parser.print_help
            parser.print_help = lambda *args, **kwargs: print_help(self.stdout)
            parser.error = perr
            if parser._subparsers:
                for subparser in (p for a in parser._subparsers._actions if isinstance(a, _SubParsersAction)
                                  for p in a.choices.values()):
                    redirect(subparser)

        self.parser = ArgumentParser(description="grommunio admin cli")
        subparsers = self.parser.add_subparsers()
        for name, handler, parserSetup, kwargs in sorted(self.funcs):
            subp = subparsers.add_parser(name, **kwargs)
            subp.set_defaults(_handle=handler)
            parserSetup(subp)
        redirect(self.parser)

    SUCCESS = 0
    ERR_DECLINE = 1
    ERR_USR_ABRT = 2

    def confirm(self, prompt=""):
        """Display confirmation dialogue.

        Parameters
        ----------
        prompt : string, optional
            Prompt to display. The default is "".

        Returns
        -------
        int
            SUCCESS if user input was "y", ERR_USR_ABRT if user aborted, ERR_DECLINE otherwise
        """
        try:
            return Cli.SUCCESS if self.input(prompt).lower() == "y" else Cli.ERR_DECLINE
        except CliError:
            raise
        except:
            return Cli.ERR_USR_ABRT

    def choice(self, prompt="", choices=(), default=None):
        """Display choice dialogue.

        If a default value is set, it is returned on empty input.

        Parameters
        ----------
        prompt : str, optional
            Prompt to display. The default is "".
        choices : list-like, optional
            Valid choices. The default is ().
        default : str, optional
            Default value. The default is None.

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        while True:
            try:
                res = self.input(prompt)
            except KeyboardInterrupt:
                return None
            if default is not None and res == "":
                return default
            if choices and res not in choices:
                continue
            return res

    @staticmethod
    def require(*args):
        """Check component availability.

        Each argument as a string containing the component name.
        If a component is not available, a CliError is raised, which is automatically caught by the Cli.
        Known components are:
            - DB: The MySQL database
            - LDAP: A connection to the LDAP server

        Parameters
        ----------
        *args : string
            Component names. Unknown components are ignored.
        """
        if "DB" in args:
            from orm import DB
            if DB is None:
                raise CliError("Database not available")
            DB.session.rollback()
            err = DB.testConnection()
            if err is not None:
                raise CliError(err)

    def input(self, prompt="", secret=False):
        """Display input prompt.

        If stdin is redirected from sys.stdin, the prompt is printed and stdin.readline is called.
        The result is automatically echoed to stdout if stdin is not a tty.

        If stdin is disabled, a CliError is raised.

        Parameters
        ----------
        prompt : str, optional
            Prompt to display. The default is "".
        secret : bool, optional
            Perform secret input (use getpass if available). The default is False.

        Raises
        ------
        CliError
            Input is not available (disabled)

        Returns
        -------
        str
            User input
        """
        if self.stdin is None:
            raise CliError("Input required but not available.")
        import sys
        if self.stdin == sys.stdin:
            if not secret:
                if self.stdout != sys.stdout:
                    self.print(prompt, end="")
                v = input(prompt if self.stdout == sys.stdout else "")
                if not self.stdin.isatty():
                    self.print(v)
                return v
            else:
                from getpass import getpass
                return getpass(prompt, self.stdout)
        else:
            self.print(prompt, end="")
            v = self.stdin.readline()
            if v.endswith("\n"):
                v = v[:-1]
            if not self.stdin.isatty() and not secret:
                self.print(v)
            return v

    def print(self, msg="", *args, **kwargs):
        """Print message to stdout.

        Provides abstraction layer for redirected stdout.

        Parameters
        ----------
        msg : str, optional
            Message to print. The default is "".
        *args : any
            Arguments forwarded to print
        **kwargs : any
            Keyword arguments forwarded to print
        """
        print(msg, *args, file=kwargs.pop("file", self.stdout), **kwargs)

    def open(self, file, mode="r", conf=False):
        """Open file.

        Provides abstraction layer for emulated filesystem. Can be used as context like builtin open function.

        If filesystem emulation is not active, the call is forwarded to builtin open().

        On emulated file systems, opening a file for reading will return the associated StringIO/BytesIO object
        if the file is provided in the emulated filesystem, otherwise a FileNotFoundError is raised.
        Opening a file for writing will create a filesystem entry and return a StringIO or BytesIO object, depending on
        whether the mode contains a 'b'.
        Closing a file on an emulated filesystem will store the content in the filesystem entry and remove the file object.

        If `conf` is set to True, the filename is interpreted as a config value containing the path.
        If the config value does not contain a path, a key error is raised.

        Parameters
        ----------
        file : str
            Name of the file
        mode : str, optional
            File open mode. The default is "r".
        conf : bool, optional
            Filename is a config value. The default is False.

        Raises
        ------
        KeyError
            `conf` is True but `file` is not a valid config value.
        FileNotFoundError
            File is opened for reading but does not exist

        Returns
        -------
        file-like
            Object providing IO interface
        """
        if conf:
            from tools.config import Config
            path = Config
            for part in file.split("."):
                path = path.get(part)
            if not isinstance(path, str):
                raise KeyError("Config value '{}' not set".format(file))
            conf = file
            file = path
        if self.fs is None:
            return open(file, mode, encoding=None if "b" in mode else "utf-8")
        if "r" in mode:
            if file not in self.fs:
                raise FileNotFoundError(2, "No such file or directory in emulated filesystem")
        elif "w" in mode:
            import io
            stream = io.BytesIO() if "b" in mode else io.StringIO()
            stream.close = self._closebuf(file)
            self.fs[file] = dict(stream=stream, mode=mode, conf=conf or None)
        return self.fs[file]["stream"]

    def _closebuf(self, file):
        """Close buffer and store contents."""
        def close():
            self.fs[file]["content"] = self.fs[file].pop("stream").getvalue()
        return close

    def closeFiles(self):
        """Close all files on emulated filesystem.

        Closes all open files and stores their content.
        Files opened in read mode are deleted.

        Does not have any effect if filesystem emulation is not active and for files in write mode that are already closed.
        """
        if self.fs is None:
            return
        for file, state in self.fs.items():
            if "stream" in state and "w" in state["mode"]:
                state["content"] = state.pop("stream").getvalue()
        readonly = tuple(file for file, state in self.fs.items() if "w" not in state["mode"])
        for f in readonly:
            self.fs.pop(f)

    def complete(self, text, state=None, maxCompletions=20):
        """Provide completions for given prefix.

        Parameters
        ----------
        text : str
            Input prefix.
        state : int, optional
            If set, provide the `state`th completion (readline mode). The default is None.
        maxCompletions : int, optional
            Maximum number of completions to return (complete mode). The default is 20.

        Returns
        -------
        str or list
            Completion or None (readline mode), or list of completions (complete mode)
        """
        self.__completing = True
        if self.completer is None:
            self.completer = argcomplete.CompletionFinder(self.parser, always_complete_options=False)
        if state is not None:
            res = self.completer.rl_complete(text, state)
            self.__completing = False
            return res
        completions = []
        for i in range(maxCompletions):
            completion = self.completer.rl_complete(text, i)
            if completion is None:
                break
            completions.append(completion)
        self.__completing = False
        return completions

    @staticmethod
    def parser_stub(parser):
        """Print usage for parser.

        Convenience function to setup a stub for a parser.
        Used for sub-parsers that do not have their own handler function.

        Parameters
        ----------
        parser : ArgumentParser
            Parser to create help stub for
        """
        parser.set_defaults(_handle=lambda *args: parser.print_usage())


from . import config, dbconf, dbtools, domain, exmdb, fetchmail, fs, ldap, mconf, misc, mlist, remote, server, services, user



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2022 grommunio GmbH

import api
from api.core import API, secure
from api.security import checkPermissions

from tools.misc import RecursiveDict
from tools.permissions import SystemAdminPermission, DomainAdminPermission, DomainAdminROPermission

from flask import jsonify, request

import json

try:
    with open("res/storelangs.json", encoding="utf-8") as file:
        storeLangs = json.load(file)
except Exception as err:
    API.logger.warn("Failed to load store languages ({}): {}"
                    .format(type(err).__name__, " - ".join(str(arg) for arg in err.args)))
    storeLangs = []


@API.route(api.BaseRoute+"/defaults/storeLangs", methods=["GET"])
@secure(requireAuth=False)
def getStoreLangs():
    return jsonify(data=storeLangs)


@API.route(api.BaseRoute+"/defaults/createParams", methods=["GET"])
@secure(requireDB=True)
def getSystemDefaults():
    checkPermissions(DomainAdminROPermission("*"))
    from orm.misc import DBConf
    data = RecursiveDict({"user": {}, "domain": {}})
    data.update(DBConf.getFile("grommunio-admin", "defaults-system", True))
    if "domain" in request.args and request.args["domain"].isdecimal:
        checkPermissions(DomainAdminROPermission(int(request.args["domain"])))
        data.update(DBConf.getFile("grommunio-admin", "defaults-domain-"+request.args["domain"]))
    return jsonify(data=data)


@API.route(api.BaseRoute+"/defaults/createParams", methods=["PATCH", "PUT"])
@secure(requireDB=True)
def setSystemDefaults():
    checkPermissions(SystemAdminPermission())
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing data"), 400
    from orm.misc import DBConf, DB
    data = RecursiveDict(data)
    if request.method == "PATCH":
        orig = DBConf.getFile("grommunio-admin", "defaults-system", True)
        data = orig.update(data)
    DBConf.setFile("grommunio-admin", "defaults-system", data)
    DB.session.commit()
    return jsonify(message="Success")


@API.route(api.BaseRoute+"/defaults/createParams/<int:domainID>", methods=["GET"])
@secure(requireDB=True)
def getDomainDefaults(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.misc import DBConf
    data = RecursiveDict({"user": {}, "domain": {}})
    data.update(DBConf.getFile("grommunio-admin", "defaults-domain-"+str(domainID), True))
    return jsonify(data=data)


@API.route(api.BaseRoute+"/defaults/createParams/<int:domainID>", methods=["PATCH", "PUT"])
@secure(requireDB=True)
def setDomainDefaults(domainID):
    checkPermissions(DomainAdminPermission(domainID))
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing data"), 400
    from orm.misc import DBConf, DB
    data = RecursiveDict(data)
    if request.method == "PATCH":
        orig = DBConf.getFile("grommunio-admin", "defaults-domain-"+str(domainID), True)
        data = orig.update(data)
    DBConf.setFile("grommunio-admin", "defaults-domain-"+str(domainID), data)
    DB.session.commit()
    return jsonify(message="Success")



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

import api
from api.core import API, secure
from api.security import checkPermissions

from datetime import datetime
from flask import jsonify, request

from tools.config import Config
from tools.logs import LogReader
from tools.permissions import SystemAdminROPermission

@API.route(api.BaseRoute+"/system/logs", methods=["GET"])
@secure()
def getLogs():
    checkPermissions(SystemAdminROPermission())
    return jsonify(data=sorted(Config["logs"]))


@API.route(api.BaseRoute+"/system/logs/<file>", methods=["GET"])
@secure()
def getLog(file):
    checkPermissions(SystemAdminROPermission())
    log = Config["logs"].get(file)
    if log is None:
        return jsonify(message="Log file not found"), 404
    n = int(request.args.get("n", 10))
    skip = int(request.args.get("skip", 0))
    after = datetime.strptime(request.args["after"], "%Y-%m-%d %H:%M:%S.%f") if "after" in request.args else None
    return jsonify(data=LogReader.tail(log.get("format", "journald"), log["source"], n, skip, after))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, InvalidUseError

from argparse import ArgumentParser


_commitHelp = "Commit configuration changes to the service. Commit actions can be configured for a service by setting "\
              "`commit_service`, `commit_file`, `commit_key` and `commit_any` parameters in `grommunio-dbconf/<service>`. "


def _autocompService(prefix, **kwargs):
    try:
        from orm.misc import DBConf
        return (entry.service for entry in
                DBConf.query.filter(DBConf.service.like(prefix+"%"))
                .with_entities(DBConf.service).all())
    except:
        return ()

def _autocompFile(prefix, parsed_args, **kwargs):
    try:
        from orm.misc import DBConf
        return (entry.file for entry in
                DBConf.query.filter(DBConf.service == parsed_args.service,
                                    DBConf.file.like(prefix+"%")).all())
    except:
        return ()

def _autocompKey(prefix, parsed_args, **kwargs):
    try:
        from orm.misc import DBConf
        return (entry.key for entry in
                DBConf.query.filter(DBConf.service == parsed_args.service,
                                    DBConf.file == parsed_args.file,
                                    DBConf.key.like(prefix+"%")).all())
    except:
        return ()

def cliDbconfSet(args):
    cli = args._cli
    cli.require("DB")
    from orm.misc import DB, DBConf
    entry = DBConf.query.filter(DBConf.service == args.service, DBConf.file == args.file, DBConf.key == args.key).first()
    if entry is None:
        entry = DBConf(service=args.service, file=args.file, key=args.key, value=args.value)
        DB.session.add(entry)
    else:
        if args.init and args.value != entry.value:
            cli.print(cli.col("Key exists - aborted.", "yellow"))
            return 1
        entry.value = args.value
    DB.session.commit()
    cli.print("{}={}".format(entry.key, entry.value or ""))
    if not args.batch:
        from tools import dbconf
        error = dbconf.commit(args.service, args.file, args.key)
        if error is not None:
            cli.print(cli.col("Commit failed: "+error, "yellow"))
            return 2


def cliDbconfGet(args):
    cli = args._cli
    cli.require("DB")
    from orm.misc import DB, DBConf
    DB.session.rollback()
    query = DBConf.query.filter(DBConf.service == args.service, DBConf.file == args.file)
    if args.key is not None:
        query = query.filter(DBConf.key == args.key)
    entries = query.with_entities(DBConf.key, DBConf.value).all()
    for entry in entries:
        cli.print("{}={}".format(entry.key, entry.value or ""))


def cliDbconfDelete(args):
    cli = args._cli
    cli.require("DB")
    from orm.misc import DB, DBConf
    query = DBConf.query.filter(DBConf.service == args.service)
    if args.file is not None:
        query = query.filter(DBConf.file == args.file)
    if args.key is not None:
        query = query.filter(DBConf.key == args.key)
    deleted = query.delete()
    DB.session.commit()
    cli.print("Deleted {} entr{}".format(deleted, "y" if deleted == 1 else "ies"))


def cliDbconfList(args):
    cli = args._cli
    cli.require("DB")
    from orm.misc import DB, DBConf
    DB.session.rollback()
    query = DBConf.query
    if args.service is None:
        header = "Services:"
        target = DBConf.service.distinct()
    else:
        header = "Files in "+args.service
        target = DBConf.file.distinct()
        query = query.filter(DBConf.service == args.service)
    if args.file is not None:
        header = "Keys in {}/{}:".format(args.service, args.file)
        target = DBConf.key.distinct()
        query = query.filter(DBConf.file == args.file)
    cli.print(header)
    entries = query.with_entities(target).all()
    if len(entries) == 0:
        cli.print(cli.col("  (no entries)", "yellow"))
    for entry in entries:
        cli.print("  "+entry[0])


def cliDbConfCommit(args):
    cli = args._cli
    cli.require("DB")
    from tools import dbconf
    error = dbconf.commit(args.service, args.file, args.key)
    if error:
        cli.print(cli.col("Commit failed: "+error, "yellow"))
        return 2


def cliDbConfCommands(args):
    def printEntries(macros, commands, prefix):
        for macro, command in macros.items():
            cli.print(prefix+cli.col(macro, attrs=[])+cli.col(" -> "+command, attrs=["dark"]))
        for command in commands:
            cli.print(prefix+cli.col(command, attrs=[]))
        cli.print()

    cli = args._cli
    prefix = "" if args.level else "  "
    from tools import dbconf
    if args.level in (None, "key"):
        if not args.level:
            cli.print(cli.col("key-level commands:", attrs=["underline"]))
        printEntries(dbconf.keyMacros, dbconf.keyCommits, prefix)
    if args.level in (None, "file"):
        if not args.level:
            cli.print(cli.col("file-level commands:", attrs=["underline"]))
        printEntries(dbconf.fileMacros, dbconf.fileCommits, prefix)
    if args.level in (None, "service"):
        if not args.level:
            cli.print(cli.col("service-level commands:", attrs=["underline"]))
        printEntries(dbconf.serviceMacros, dbconf.serviceCommits, prefix)


def _setupCliDbconf(subp: ArgumentParser):
    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    commands = sub.add_parser("commands", help="Show whitelisted commit commands")
    commands.set_defaults(_handle=cliDbConfCommands)
    commands.add_argument("level", nargs="?", choices=("key", "file", "service"), help="Show only commands for specific level")
    commit = sub.add_parser("commit", description=_commitHelp, help="Run commit function")
    commit.set_defaults(_handle=cliDbConfCommit)
    commit.add_argument("service", help="Service to update").completer = _autocompService
    commit.add_argument("file", nargs="?", help="File to commit. If omittet, commit all files").completer = _autocompFile
    commit.add_argument("key", nargs="?", help="Key to commit. If omittet, commit all keys").completer = _autocompKey
    delete = sub.add_parser("delete", help="Delete service, file or key")
    delete.set_defaults(_handle=cliDbconfDelete)
    delete.add_argument("service", help="Service to delete from").completer = _autocompService
    delete.add_argument("file", nargs="?", help="File to delete from. If omitted, delete all files").completer = _autocompFile
    delete.add_argument("key", nargs="?", help="Configuration entry to delete. If omitted, delete all entries")\
        .completer = _autocompKey
    get = sub.add_parser("get", help="Print file or key")
    get.set_defaults(_handle=cliDbconfGet)
    get.add_argument("service", help="Service get configuration for").completer = _autocompService
    get.add_argument("file", help="File or section. If omitted print all files").completer = _autocompFile
    get.add_argument("key", nargs="?", help="Configuration entry. If omitted, print complete file").completer = _autocompKey
    list = sub.add_parser("list", help="List services, files or keys")
    list.set_defaults(_handle=cliDbconfList)
    list.add_argument("service", nargs="?", help="Service to list files from").completer = _autocompService
    list.add_argument("file", nargs="?", help="File to list keys from").completer = _autocompFile
    set = sub.add_parser("set", help="Set configuration value")
    set.set_defaults(_handle=cliDbconfSet)
    set.add_argument("service", help="Service to configure").completer = _autocompService
    set.add_argument("file", help="File or section").completer = _autocompFile
    set.add_argument("key", help="Configuration entry").completer = _autocompKey
    set.add_argument("value", nargs="?", help="Value to set. If omitted, create an empty entry")
    set.add_argument("-b", "--batch", action="store_true", help="Do not autocommit changes")
    set.add_argument("-i", "--init", action="store_true", help="Only set if value configuration does not exist")


@Cli.command("dbconf", _setupCliDbconf, help="Database-stored configuration management")
def cliDbconfStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

import api
from api.core import API, secure
from api.security import checkPermissions

from flask import request, jsonify

from .. import defaultListHandler, defaultObjectHandler

from tools.permissions import DomainAdminPermission, DomainAdminROPermission

from orm import DB


@API.route(api.BaseRoute+"/domains/<int:domainID>/mlists", methods=["GET"])
@secure(requireDB=True)
def mlistListEndpoint(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.mlists import MLists
    return defaultListHandler(MLists, (MLists.domainID == domainID,))


@API.route(api.BaseRoute+"/domains/<int:domainID>/mlists", methods=["POST"])
@secure(requireDB=True)
def createMlist(domainID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.mlists import MLists
    data = request.get_json(silent=True, cache=True) or {}
    data["domainID"] = domainID
    return defaultListHandler(MLists)


@API.route(api.BaseRoute+"/domains/<int:domainID>/mlists/<int:ID>", methods=["GET", "PATCH", "DELETE"])
@secure(requireDB=True)
def mlistObjectEndpoint(domainID, ID):
    checkPermissions(DomainAdminROPermission(domainID) if request.method == "GET" else DomainAdminPermission(domainID))
    from orm.mlists import MLists
    return defaultObjectHandler(MLists, ID, "Mailing list", filters=(MLists.domainID == domainID,))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from math import ceil
import os
import shutil
import subprocess

from .misc import setDirectoryOwner, setDirectoryPermission
from .structures import XID, GUID
from .config import Config
from .constants import PropTags, ConfigIDs, PublicFIDs, PrivateFIDs, Misc
from .rop import ntTime

import traceback

import sqlite3
import time

import logging
logger = logging.getLogger("storage")


def genPath(index: int, depth: int):
    """Generate minimum width unique path for index.

    Parameters
    ----------
    index : int
        Index of the file
    depth : int
        Number of levels.

    Returns
    -------
    list of int
        Indices of each directory
    """
    def minSplits(num: int, depth: int):
        """Calculate minimum number of splits for `num` items in a tree with `depth` levels.

        Parameters
        ----------
        num : int
            Number of items in the tree.
        depth : int
            Depth of the tree

        Returns
        -------
        int
            Minimal required splits at any level.
        """
        return ceil(num**(1/depth))

    diridx = [0]*depth  # Path
    ms = minSplits(index+1, depth)  # Minimum required splits
    splitidx = index - (ms-1)**depth  # Index within current split level
    subprev = index  # Only required for depth == 1
    created = False  # Whether in a new branch (otherwise filling an old one)
    for level in range(depth-1):  # Traverse tree down to leaves
        subcap = ms**(depth-level-1)  # Capacity of each sub tree
        subprev = (ms-1)**(depth-level-1)  # Sub tree capacity at previous split level
        subnew = subcap if created else subcap-subprev  # Number of new elements that can be placed in each sub tree
        diridx[level] = min(splitidx//subnew, ms-1)  # Choose split number according to capacity
        splitidx -= diridx[level]*subnew  # Calculate local index for subtree
        created |= diridx[level] == ms-1  # Set sticky fresh-branch flag
    diridx[-1] = splitidx if created else splitidx+subprev  # Set leaf index
    return diridx


def createPath(parent: str, index: int, depth: int, fileUid=None, fileGid=None):
    """Create storage path.

    Parameters
    ----------
    parent : str
        Parent directory
    index : int
        Index of the element to store
    depth : int
        Number of levels to create

    Raises
    ------
    OSError
        Directory creation failed.

    Returns
    -------
    path : str
        The full path of the created directory (without trailing slash)
    """
    subdirs = ["{:X}".format(i) for i in genPath(index, depth)]
    path = os.path.join(parent, *subdirs)
    os.makedirs(path)
    if fileUid is not None or fileGid is not None:
        temp = parent
        for subdir in subdirs:
            temp = os.path.join(temp, subdir)
            try:
                shutil.chown(temp, fileUid, fileGid)
            except Exception:
                pass
    return path


class SetupContext:
    def __enter__(self):
        """Enter context."""
        self._dirs = []
        self.success = False
        return self

    def __exit__(self, *args):
        """Exit context.

        If success is not set to True, any directories created are removed.
        """
        if not self.success:
            for d in self._dirs:
                try:
                    shutil.rmtree(d)
                except Exception:
                    pass
        if getattr(self, "exmdb", None) is not None:
            self.exmdb.rollback()

    def createGenericFolder(self, folderID: int, objectID: int):
        """Create a generic MS Exchange folder.

        Parameters
        ----------
        folderID : int
            ID of the new folder.
        parentID : int
            ID of the parent folder (or `None` to create root folder).
        objectID : int
            ID of the domain to create the folder for.
        displayName : str
            Name of the folder.
        containerClass : str, optional
            Container class of the folder. The default is None.
        """
        currentEid = self.lastEid+1
        self.lastEid += Misc.ALLOCATED_EID_RANGE
        self.exmdb.execute("INSERT INTO allocated_eids VALUES (?, ?, ?, 1)", (currentEid, self.lastEid, int(time.time())))
        self.lastCn += 1
        self.lastArt += 1
        ntNow = ntTime()
        xidData = XID.fromDomainID(objectID, self.lastCn).serialize()
        stmt = "INSERT INTO folder_properties VALUES (?, ?, ?)"
        self.exmdb.execute(stmt, (folderID, PropTags.CREATIONTIME, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.LASTMODIFICATIONTIME, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.LOCALCOMMITTIMEMAX, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.HIERREV, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.CHANGEKEY, xidData))
        self.exmdb.execute(stmt, (folderID, PropTags.PREDECESSORCHANGELIST, b'\x16'+xidData))

    def mkext(self, command, name):
        """Try to databases with external tools.

        Executes shell command to create database files.
        Fails if the command terminates with non-zero exit code.

        Parameters
        ----------
        command : str
            Command to execute
        name : str
            Name of the entity.

        Returns
        -------
        bool
            True if successful, False otherwise.
        """
        try:
            res = subprocess.run((command, name), stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
            if res.returncode:
                logger.warning("{} return non-zero exit code ({}): {}".format(command, res.returncode, res.stdout))
                return False
            if res.stdout:
                logger.debug("{} (stdout): {}".format(command, res.stdout))
            if res.stderr:
                logger.debug("{} (stderr): {}".format(command, res.stderr))
        except Exception as err:
            logger.error("Failed to run {} ({}): {}".format(command, type(err).__name__,
                                                            " - ".join(str(arg) for arg in err.args)))
            return False
        return True


class DomainSetup(SetupContext):
    """Domain initialization context.

    Can be used in a with context to ensure automatic cleanup of created directories if an error occurs.
    Invocation of the run() method creates the complete directory structure required by a domain and
    sets up the initial MS Exchange database.

    If everything went well, the `success` attribute is set to True.
    If any exception occurs it is caught and the stack trace is written to the log. In this case, the `error` attribute
    contains a short error description and the `errorCode` attribute is set to an appropriate HTTP status code.
    """

    def __init__(self, domain, session):
        """Initialize context object

        Parameters
        ----------
        domain : orm.domains.Domains
            Domain to initialize.
        """

        self.lastEid = Misc.ALLOCATED_EID_RANGE
        self.lastCn = Misc.CHANGE_NUMBER_BEGIN
        self.lastArt = 0

        self.domain = domain
        self.session = session

        self.success = False
        self.error = self.errorCode = None

    def run(self):
        """Run domain home directory initialization."""
        try:
            fileUid, fileGid = Config["options"].get("fileUid"), Config["options"].get("fileGid")
            self.createHomedir(fileUid, fileGid)
            self.session.commit()
            self.createExmdb()
            try:
                setDirectoryOwner(self.domain.homedir, fileUid, fileGid)
                setDirectoryPermission(self.domain.homedir, Config["options"].get("filePermissions"))
            except Exception as err:
                logger.warn("Could not set domain directory ownership: "+" - ".join(str(arg) for arg in err.args))
            self.success = True
        except PermissionError as err:
            logger.error(traceback.format_exc())
            self.error = "Could not create home directory ({})".format(err.args[1])
            self.errorCode = 500
            self.domain.homedir = ""
        except FileExistsError:
            logger.error("Failed to create {}: Directory exists.".format(self.domain.homedir))
            self.error = "Could not create home directory: File exists"
            self.errorCode = 500
            self.domain.homedir = ""
        except Exception:
            logger.error(traceback.format_exc())
            self.error = "Unknown error"
            self.errorCode = 500
            self.domain.homedir = ""

    def createHomedir(self, fileUid, fileGid):
        """Set up directory structure for a domain.

        Creates the home directory according to its ID in the prefix specified in the configuration.
        Intermediate directories are created automatically if necessary.

        Additional `cid`, `log` and `tmp` subdirectories are created in the home directory.
        """
        options = Config["options"]
        self.domain.homedir = createPath(self.domain.homedir, self.domain.ID, options["domainStorageLevels"], fileUid, fileGid)
        self._dirs.append(self.domain.homedir)
        if options["domainAcceleratedStorage"] is not None:
            dbPath = createPath(options["domainAcceleratedStorage"], self.domain.ID, options["domainStorageLevels"])
            self._dirs.append(dbPath)
            os.symlink(dbPath, self.domain.homedir+"/exmdb")
        else:
            os.mkdir(self.domain.homedir+"/exmdb")
        os.mkdir(self.domain.homedir+"/cid")
        os.mkdir(self.domain.homedir+"/log")
        os.mkdir(self.domain.homedir+"/tmp")

    def createExmdb(self):
        """Create exchange SQLite database for domain.

        Database is placed under <homedir>/exmdb/exchange.sqlite3.
        """
        if self.mkext("gromox-mkpublic", self.domain.domainname):
            return
        dbPath = os.path.join(self.domain.homedir, "exmdb", "exchange.sqlite3")
        shutil.copy("res/domain.sqlite3", dbPath)
        self.exmdb = sqlite3.connect(dbPath)
        self.exmdb.execute("INSERT INTO store_properties VALUES (?, ?)", (PropTags.CREATIONTIME, ntTime()))
        self.createGenericFolder(PublicFIDs.ROOT, self.domain.ID)
        self.createGenericFolder(PublicFIDs.IPMSUBTREE, self.domain.ID)
        self.createGenericFolder(PublicFIDs.NONIPMSUBTREE, self.domain.ID)
        self.createGenericFolder(PublicFIDs.EFORMSREGISTRY, self.domain.ID)
        self.exmdb.execute("INSERT INTO configurations VALUES (?, ?)", (ConfigIDs.MAILBOX_GUID, str(GUID.random())))
        self.exmdb.commit()
        self.exmdb.close()
        self.exmdb = None


class UserSetup(SetupContext):
    """User initialization context.

    Can be used in a with context to ensure automatic cleanup of created directories if an error occurs.
    Invocation of the run() method creates the complete directory structure required by a user and
    sets up the initial MS Exchange databases (exchange.sqlite3 and midb.sqlite3).

    If everything went well, the `success` attribute is set to True.
    If any exception occurs it is caught and the stack trace is written to the log. In this case, the `error` attribute
    contains a short error description and the `errorCode` attribute is set to an appropriate HTTP status code.
    """

    def __init__(self, user, session):
        """Initialize context object.

        Parameters
        ----------
        user : orm.users.Users
            User to initialize.
        """
        self.lastEid = Misc.ALLOCATED_EID_RANGE
        self.lastCn = Misc.CHANGE_NUMBER_BEGIN
        self.lastArt = 0

        self.user = user
        self.session = session

        self.success = False
        self.error = self.errorCode = None

    def run(self):
        """Run user home directory initialization."""
        try:
            fileUid, fileGid = Config["options"].get("fileUid"), Config["options"].get("fileGid")
            self.createHomedir(fileUid, fileGid)
            self.session.commit()
            self.createExmdb()
            self.createMidb()
            try:
                setDirectoryOwner(self.user.maildir, fileUid, fileGid)
                setDirectoryPermission(self.user.maildir, Config["options"].get("filePermissions"))
            except Exception as err:
                logger.warn("Could not set user directory ownership: "+" - ".join(str(arg) for arg in err.args))
            self.success = True
        except PermissionError as err:
            logger.error(traceback.format_exc())
            self.error = "Could not create home directory ({})".format(err.args[1])
            self.errorCode = 500
            self.user.maildir = ""
        except FileExistsError:
            logger.error("Failed to create {}: Directory exists.".format(self.domain.homedir))
            self.error = "Could not create home directory: File exists"
            self.errorCode = 500
            self.user.maildir = ""
        except Exception:
            logger.error(traceback.format_exc())
            self.error = "Unknown error"
            self.errorCode = 500
            self.user.maildir = ""

    def createHomedir(self, fileUid=None, fileGid=None):
        """Set up directory structure for a user.

        Creates the home directory according to its ID in the prefix set in the configuration.
        Intermediate directories are created automatically if necessary.

        If `userAcceleratedStorage` is set, an exmdb path is created and symlinked accordingly.

        Additional `cid`, `config`, `eml`, `ext` and `tmp` subdirectories are created in the home directory.
        """
        options = Config["options"]
        self.user.maildir = createPath(self.user.maildir, self.user.ID, options["userStorageLevels"], fileUid, fileGid)
        self._dirs.append(self.user.maildir)
        if options["userAcceleratedStorage"] is not None:
            dbPath = createPath(options["userAcceleratedStorage"], self.user.ID, options["userStorageLevels"])
            self._dirs.append(dbPath)
            os.symlink(dbPath, self.user.maildir+"/exmdb")
        else:
            os.mkdir(self.user.maildir+"/exmdb")
        os.mkdir(self.user.maildir+"/tmp")
        os.mkdir(self.user.maildir+"/tmp/imap.rfc822")
        os.mkdir(self.user.maildir+"/tmp/faststream")
        os.mkdir(self.user.maildir+"/eml")
        os.mkdir(self.user.maildir+"/ext")
        os.mkdir(self.user.maildir+"/cid")
        os.mkdir(self.user.maildir+"/config")
        thumbnailSrc = os.path.join(Config["options"]["dataPath"], Config["options"]["portrait"])
        try:
            shutil.copy(thumbnailSrc, self.user.maildir+"/config/portrait.jpg")
        except FileNotFoundError:
            pass

    def createSearchFolder(self, folderID: int, userID: int):
        """Create exmdb search folder entries."""
        self.lastCn += 1
        self.lastArt += 1
        ntNow = ntTime()
        xidData = XID.fromDomainID(userID, self.lastCn).serialize()
        stmt = "INSERT INTO folder_properties VALUES (?,?,?)"
        self.exmdb.execute(stmt, (folderID, PropTags.CREATIONTIME, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.LASTMODIFICATIONTIME, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.HIERREV, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.LOCALCOMMITTIMEMAX, ntNow))
        self.exmdb.execute(stmt, (folderID, PropTags.CHANGEKEY, xidData))
        self.exmdb.execute(stmt, (folderID, PropTags.PREDECESSORCHANGELIST, b'\x16'+xidData))

    def createExmdb(self):
        """Create exchange SQLite database for user.

        Database is placed under <homedir>/exmdb/exchange.sqlite3.
        """
        if self.mkext("gromox-mkprivate", self.user.username):
            return
        dbPath = os.path.join(self.user.maildir, "exmdb", "exchange.sqlite3")
        shutil.copy("res/user.sqlite3", dbPath)
        self.exmdb = sqlite3.connect(dbPath)
        ntNow = ntTime()
        stmt = "INSERT INTO receive_table VALUES (?, ?, ?)"
        self.exmdb.execute(stmt, ("", PrivateFIDs.INBOX, ntNow))
        self.exmdb.execute(stmt, ("IPC", PrivateFIDs.ROOT, ntNow))
        self.exmdb.execute(stmt, ("IPM", PrivateFIDs.INBOX, ntNow))
        self.exmdb.execute(stmt, ("REPORT.IPM", PrivateFIDs.INBOX, ntNow))
        self.createGenericFolder(PrivateFIDs.ROOT, self.user.ID)
        self.createGenericFolder(PrivateFIDs.IPMSUBTREE, self.user.ID)
        self.createGenericFolder(PrivateFIDs.INBOX, self.user.ID)
        self.createGenericFolder(PrivateFIDs.DRAFT, self.user.ID)
        self.createGenericFolder(PrivateFIDs.OUTBOX, self.user.ID)
        self.createGenericFolder(PrivateFIDs.SENT_ITEMS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.DELETED_ITEMS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.CONTACTS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.CALENDAR, self.user.ID)
        self.createGenericFolder(PrivateFIDs.JOURNAL, self.user.ID)
        self.createGenericFolder(PrivateFIDs.NOTES, self.user.ID)
        self.createGenericFolder(PrivateFIDs.TASKS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.QUICKCONTACTS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.IMCONTACTLIST, self.user.ID)
        self.createGenericFolder(PrivateFIDs.GALCONTACTS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.JUNK, self.user.ID)
        self.createGenericFolder(PrivateFIDs.CONVERSATION_ACTION_SETTINGS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.DEFERRED_ACTION, self.user.ID)
        self.createSearchFolder(PrivateFIDs.SPOOLER_QUEUE, self.user.ID)
        self.createGenericFolder(PrivateFIDs.COMMON_VIEWS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.SCHEDULE, self.user.ID)
        self.createGenericFolder(PrivateFIDs.FINDER, self.user.ID)
        self.createGenericFolder(PrivateFIDs.VIEWS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.SHORTCUTS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.SYNC_ISSUES, self.user.ID)
        self.createGenericFolder(PrivateFIDs.CONFLICTS, self.user.ID)
        self.createGenericFolder(PrivateFIDs.LOCAL_FAILURES, self.user.ID)
        self.createGenericFolder(PrivateFIDs.SERVER_FAILURES, self.user.ID)
        self.createGenericFolder(PrivateFIDs.LOCAL_FREEBUSY, self.user.ID)
        self.exmdb.execute("INSERT INTO configurations VALUES (?, ?)", (ConfigIDs.MAILBOX_GUID, str(GUID.random())))
        self.exmdb.commit()
        self.exmdb.close()
        self.exmdb = None

    def createMidb(self):
        """Create midb SQLite database for user.

        Database is placed under <homedir>/exmdb/midb.sqlite3.
        """
        if self.mkext("gromox-mkmidb", self.user.username):
            return
        dbPath = os.path.join(self.user.maildir, "exmdb", "midb.sqlite3")
        shutil.copy("res/midb.sqlite3", dbPath)
        DB = sqlite3.connect(dbPath)
        DB.execute("INSERT INTO configurations VALUES (1, ?)", (self.user.username,))
        DB.commit()
        DB.close()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020-2021 grommunio GmbH

import api
from api.core import API, secure
from api.security import checkPermissions

from flask import request, jsonify

from services import Service

from tools.constants import PropTags, PropTypes, ExchangeErrors, PublicFIDs
from tools.permissions import DomainAdminPermission, DomainAdminROPermission
from tools.rop import nxTime, makeEidEx
from tools.tasq import TasQServer

from datetime import datetime


def folderToDict(folder):
    return {"folderid": str(folder.folderId),
            "displayname": folder.displayName,
            "comment": folder.comment,
            "creationtime": datetime.fromtimestamp(nxTime(folder.creationTime)).strftime("%Y-%m-%d %H:%M:%S"),
            "container": folder.container,
            "syncMobile": folder.syncToMobile}


def setPublicFolderMobileSync(exmdb, client, domainname, folderId, value):
    synctomobile = client.resolveNamedProperties(True, [exmdb.PropertyName(exmdb.GUID.PSETID_GROMOX, "synctomobile")])[0]
    synctomobile = (synctomobile << 16) | PropTypes.BYTE
    if not synctomobile:
        return
    problems = client.setFolderProperties(0, folderId, [exmdb.TaggedPropval(synctomobile, value)])
    with Service("redis") as redis:
        redis.delete("grommunio-sync:sharedfolders-"+domainname)
    return problems


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders", methods=["GET"])
@secure(requireDB=True)
def getPublicFoldersList(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    limit = int(request.args.get("limit", 50))
    offset = int(request.args.get("offset", 0))
    parent = int(request.args.get("parentID", makeEidEx(1, PublicFIDs.IPMSUBTREE)))
    with Service("exmdb") as exmdb:
        if "match" in request.args:
            fuzzyLevel = exmdb.Restriction.FL_SUBSTRING | exmdb.Restriction.FL_IGNORECASE
            match = request.args["match"]
            restriction = exmdb.Restriction.OR([
                exmdb.Restriction.CONTENT(fuzzyLevel, 0, exmdb.TaggedPropval(PropTags.DISPLAYNAME, match)),
                exmdb.Restriction.CONTENT(fuzzyLevel, 0, exmdb.TaggedPropval(PropTags.COMMENT, match))])
        else:
            restriction = exmdb.Restriction.NULL()
        client = exmdb.domain(domain)
        synctomobile = client.resolveNamedProperties(True, [exmdb.PropertyName(exmdb.GUID.PSETID_GROMOX, "synctomobile")])[0]
        synctomobile = (synctomobile << 16) | PropTypes.BYTE
        response = exmdb.FolderList(client.listFolders(parent, limit=limit, offset=offset, restriction=restriction,
                                                       proptags=client.defaultFolderProps+[synctomobile]), synctomobile)
    folders = [folderToDict(entry) for entry in response.folders]
    return jsonify(data=folders)


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/tree", methods=["GET"])
@secure(requireDB=True)
def getFolderTree(domainID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    parentID = int(request.args.get("folderID", makeEidEx(1, PublicFIDs.IPMSUBTREE)))
    tags = (PropTags.FOLDERID, PropTags.PARENTFOLDERID, PropTags.DISPLAYNAME, PropTags.CONTAINERCLASS)
    with Service("exmdb") as exmdb:
        client = exmdb.domain(domain)
        parent = exmdb.Folder(client.getFolderProperties(0, parentID, tags))
        folders = exmdb.FolderList(client.listFolders(parentID, True, tags))
    idmap = {folder.folderId: {"folderid": str(folder.folderId), "name": folder.displayName, "container": folder.container}
             for folder in folders.folders}
    idmap[parentID] = {"folderid": str(parent.folderId), "name": parent.displayName}
    for folder in folders.folders:
        parentFolder = idmap[folder.parentId]
        if "children" not in parentFolder:
            parentFolder["children"] = []
        parentFolder["children"].append(idmap[folder.folderId])
    return jsonify(idmap[parentID])


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders", methods=["POST"])
@secure(requireDB=True)
def createPublicFolder(domainID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    data = request.json
    parentID = int(data.get("parentID") or makeEidEx(1, PublicFIDs.IPMSUBTREE))
    with Service("exmdb") as exmdb:
        client = exmdb.domain(domain)
        folderId = client.createFolder(domain.ID, data["displayname"], data["container"], data["comment"], parentID)
        if folderId == 0:
            return jsonify(message="Folder creation failed"), 500
        if "syncMobile" in data:
            setPublicFolderMobileSync(exmdb, client, domain.domainname, folderId, data["syncMobile"])
    return jsonify(folderid=str(folderId),
                   displayname=data["displayname"],
                   comment=data["comment"],
                   creationtime=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                   container=data["container"],
                   syncMobile=data.get("syncMobile")), 201


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/<int:folderID>", methods=["GET"])
@secure(requireDB=True)
def getPublicFolder(domainID, folderID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    with Service("exmdb") as exmdb:
        client = exmdb.domain(domain)
        synctomobile = client.resolveNamedProperties(True, [exmdb.PropertyName(exmdb.GUID.PSETID_GROMOX, "synctomobile")])[0]
        synctomobile = (synctomobile << 16) | PropTypes.BYTE
        response = exmdb.Folder(client.getFolderProperties(0, folderID, client.defaultFolderProps+[synctomobile]),
                                synctomobile)
    return jsonify(folderToDict(response))


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/<int:folderID>", methods=["PATCH"])
@secure(requireDB=True)
def updatePublicFolder(domainID, folderID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    data = request.get_json(silent=True) or {}
    supported = ((PropTags.COMMENT, "comment"), (PropTags.DISPLAYNAME, "displayname"), (PropTags.CONTAINERCLASS, "container"))
    with Service("exmdb") as exmdb:
        proptags = [exmdb.TaggedPropval(tag, data[tagname]) for tag, tagname in supported if tagname in data]
        if not len(proptags) and "syncMobile" not in data:
            return jsonify(message="Nothing to do")
        client = exmdb.domain(domain)
        problems = client.setFolderProperties(0, folderID, proptags)
        if "syncMobile" in data:
            problems += setPublicFolderMobileSync(exmdb, client, domain.domainname, folderID, data["syncMobile"])
        if len(problems):
            errors = ["{} ({})".format(PropTags.lookup(problem.proptag, hex(problem.proptag)).lower(),
                                       ExchangeErrors.lookup(problem.err, hex(problem.err))) for problem in problems]
            return jsonify(message="Update failed for tag{} {}".format("" if len(errors) == 1 else "s",
                                                                       ", ".join(errors))), 500
    return jsonify(message="Success.")


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/<int:folderID>", methods=["DELETE"])
@secure(requireDB=True)
def deletePublicFolder(domainID, folderID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    task = TasQServer.mktask.deleteFolder(domain.homedir, folderID, False, request.args.get("clear") == "true",
                                          DomainAdminROPermission(domainID), domain.homeserver)
    timeout = float(request.args.get("timeout", 1))
    if timeout > 0:
        TasQServer.wait(task.ID, timeout)
    if not task.done:
        return jsonify(message="Created background task #"+str(task.ID), taskID=task.ID), 202
    if task.state == task.COMPLETED:
        return jsonify(message="Success")
    return jsonify(message="Folder deletion failed: "+task.message), 500


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/<int:folderID>/owners", methods=["GET"])
@secure(requireDB=True)
def getPublicFolderOwnerList(domainID, folderID):
    checkPermissions(DomainAdminROPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    with Service("exmdb") as exmdb:
        client = exmdb.domain(domain)
        response = exmdb.FolderMemberList(client.getFolderMemberList(folderID))
    owners = [{"memberID": member.id, "displayName": member.name, "username": member.mail, "permissions": member.rights}
              for member in response.members if not member.special]
    return jsonify(data=owners)


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/<int:folderID>/owners", methods=["POST"])
@secure(requireDB=True)
def addPublicFolderOwner(domainID, folderID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.domains import Domains
    data = request.get_json(silent=True)
    if data is None or "username" not in data:
        return jsonify(message="Missing required parameter 'username'"), 400
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    with Service("exmdb") as exmdb:
        client = client = exmdb.domain(domain)
        client.setFolderMember(folderID, data["username"], data.get("permissions", client.ownerRights))
    with Service("redis") as redis:
        redis.delete("grommunio-sync:sharedfolders-"+domain.domainname)
    return jsonify(message="Success"), 201


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/<int:folderID>/owners/<int:memberID>", methods=["PUT"])
@secure(requireDB=True)
def updatePublicFolderOwner(domainID, folderID, memberID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    data = request.get_json(silent=True)
    if data is None or "permissions" not in data:
        return jsonify(message="Missing required data parameter 'permissions'"), 400
    with Service("exmdb") as exmdb:
        client = exmdb.domain(domain)
        client.setFolderMember(folderID, memberID, data["permissions"], client.SET)
    with Service("redis") as redis:
        redis.delete("grommunio-sync:sharedfolders-"+domain.domainname)
    return jsonify(message="Success")


@API.route(api.BaseRoute+"/domains/<int:domainID>/folders/<int:folderID>/owners/<int:memberID>", methods=["DELETE"])
@secure(requireDB=True)
def deletePublicFolderOwner(domainID, folderID, memberID):
    checkPermissions(DomainAdminPermission(domainID))
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.ID == domainID).first()
    if domain is None:
        return jsonify(message="Domain not found"), 404
    with Service("exmdb") as exmdb:
        client = exmdb.domain(domain)
        client.setFolderMember(folderID, memberID, 0xFFFFFFFF, client.REMOVE)
    with Service("redis") as redis:
        redis.delete("grommunio-sync:sharedfolders-"+domain.domainname)
    return jsonify(message="Success"), 200



# -*- coding: utf-8 -*-

import re

class ClassFilter:
    class Condition:
        sqlops = {"eq": "=",
                  "ne": "<>",
                  "lt": "<",
                  "le": "<=",
                  "gt": ">",
                  "ge": ">=",
                  "li": "RLIKE",
                  "ex": "IS NOT NULL",
                  "nx": "IS NULL"}

        unary = {"ex", "nx"}

        columns = {"username"}

        def __init__(self, data):
            if "prop" not in data:
                raise ValueError("Missing property")
            self.type = "c" if data["prop"] in self.columns else "p"
            if self.type is None:
                raise ValueError("Cannot deduct type (missing 'c' or 'p' property)")
            self.op = data.get("op")
            self.target = data["prop"]
            if self.type == "p" and type(self.target) != int:
                raise ValueError("Invalid property")
            if self.op not in self.unary:
                if "val" not in data or data["val"] is None:
                    raise ValueError("Missing target value")
                self.value = data["val"]
            if self.op is None:
                raise ValueError("Missing operator".format(self.op))
            if self.op not in self.sqlops:
                raise ValueError("Invalid operator '{}'".format(self.op))
            if self.type == "c" and self.target not in self.columns:
                raise ValueError("Invalid column '{}'".format(self.target))
            if self.op not in self.unary and ("val" not in data or type(data["val"]) != str):
                raise ValueError("Invalid filter value (must be string)")

        def sql(self, alias):
            value = None if self.value is None else "'{}'".format(self.value if self.op == "li" else self.value)
            if self.type == "c":
                return "{}.{} {} {}".format(alias, self.target, self.sqlops[self.op], "" if self.op in self.unary else value)
            elif self.type == "p":
                return "{}.propval_str {} {}".format(alias, self.sqlops[self.op], "" if self.op in self.unary else value)

    def __init__(self, data):
        if isinstance(data, str):
            import json
            data = json.loads(data)
        self.expressions = [[self.Condition(entry) for entry in conj] for conj in data]
        if len(self.expressions) == 0 or min(len(disj) for disj in self.expressions) == 0:
            raise ValueError("Cannot use empty filter expression")

    def sql(self, columns):
        tags = {expr.target for conj in self.expressions for expr in conj if expr.type == "p"}
        joins = " ".join("LEFT JOIN user_properties AS up_{tag} on u.id=up_{tag}.user_id AND up_{tag}.proptag='{tag}'"
                          .format(tag=tag) for tag in tags)
        filters = ") AND (".join(" OR ".join(expr.sql("u" if expr.type == "c" else "up_"+str(expr.target)) for expr in conj)
                               for conj in self.expressions)
        return "SELECT {} FROM users AS u {} WHERE ({})".format(columns, joins, filters)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2022 grommunio GmbH

from . import Cli, InvalidUseError, ArgumentParser
from .common import proptagCompleter

_perms = {
    "readany": 0x1,
    "create": 0x2,
    "sendas": 0x4,
    "editowned": 0x8,
    "deleteowned": 0x10,
    "editany": 0x20,
    "deleteany": 0x40,
    "createsubfolder": 0x80,
    "folderowner": 0x100,
    "foldercontact": 0x200,
    "foldervisible": 0x400,
    "freebusysimple": 0x800,
    "freebusydetailed": 0x1000,
    "storeowner": 0x2000}
_permsAll = 0x27ff


def _getClient(args, exmdb):
    cli = args._cli
    if "@" in args.target:
        from orm.users import Users
        user = Users.query.filter(Users.username == args.target).first()
        if user is None:
            cli.print(cli.col("No user matching '{}'.".format(args.target), "red"))
            return 1, None
        client = exmdb.user(user)
        client.accountID = user.ID
        return 0, client
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.domainname == args.target).first()
    if domain is None:
        cli.print(cli.col("No domain matching '{}'.".format(args.target), "red"))
        return 1, None
    client = exmdb.domain(domain)
    client.accountID = domain.ID
    return 0, client


def _isPrivate(args):
    return "@" in args.target


class _FolderNode():
    I = chr(0x2502)+" "
    L = chr(0x2514)+chr(0x2500)
    T = chr(0x251c)+chr(0x2500)

    def __init__(self, folder, subfolders=[]):
        from tools.rop import gcToValue
        self.ID = gcToValue(folder.folderId)
        self.parentID = gcToValue(folder.parentId)
        self.name = folder.displayName
        self.subfolders = []
        if subfolders:
            self._buildTree(subfolders)

    def _buildTree(self, subfolders):
        subfolders = [f if isinstance(f, _FolderNode) else _FolderNode(f) for f in subfolders]
        idmap = {sub.ID: sub for sub in subfolders}
        idmap[self.ID] = self
        for sub in subfolders:
            if sub.parentID in idmap:
                idmap[sub.parentID].subfolders.append(sub)

    @property
    def idstr(self):
        return hex(self.ID)

    def _toDict(self, recursive=True):
        me = dict(ID=self.ID, parentID=self.parentID, name=self.name)
        if recursive:
            me["subfolders"] = [sf._toDict(True) for sf in self.subfolders]
        return me

    def _collectSubfolders(self):
        sfs = list(self.subfolders)
        for sf in self.subfolders:
            sfs += sf._collectSubfolders()
        return sfs

    def _print_json(self, flat=False):
        import json
        me = self._toDict(not flat)
        if flat:
            me["subfolders"] = [sf._toDict(False) for sf in self._collectSubfolders()]
        return json.dumps(me, separators=(",", ":"))

    def _print_pretty(self, cli, level=-1, pref=""):
        content = "{} ({})\n".format(cli.col(self.name, attrs=["bold"]), self.idstr)
        if self.subfolders:
            for sub in self.subfolders[:-1]:
                content += pref+self.T+sub._print_pretty(cli, level+1, pref+self.I)
            content += pref+self.L+self.subfolders[-1]._print_pretty(cli, level+1, pref+"  ")
        return content if level >= 0 else content[:-1]

    def print(self, cli, format="pretty"):
        if format in ("json-flat", "json-tree"):
            return self._print_json(format == "json-flat")
        return self._print_pretty(cli)

    def tabledata(self):
        sfs = self._collectSubfolders()
        return [(self.ID, self.parentID, self.name)]+[(sf.ID, sf.parentID, sf.name) for sf in sfs]


def cliExmdbFolderCreate(args):
    cli = args._cli
    cli.require("DB")
    from services import Service
    from tools.constants import PublicFIDs, PrivateFIDs
    from tools.rop import makeEidEx
    parentID = args.ID or (PrivateFIDs.IPMSUBTREE if _isPrivate(args) else PublicFIDs.IPMSUBTREE)
    parentID = makeEidEx(1, parentID)
    with Service("exmdb") as exmdb:
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        folderID = client.createFolder(client.accountID, args.name, args.type, args.comment, parentID)
        if folderID == 0:
            cli.print(cli.col("Folder creation failed", "red"))
            return 2
        folder = exmdb.Folder(client.getFolderProperties(0, folderID, client.defaultFolderProps))
        cli.print(_FolderNode(folder).print(cli))


def cliExmdbFolderDelete(args):
    cli = args._cli
    cli.require("DB")
    from services import Service
    from tools.constants import PublicFIDs, PrivateFIDs
    from tools.rop import makeEidEx, gcToValue
    with Service("exmdb") as exmdb:
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        try:
            fids = [makeEidEx(1, int(args.folderspec, 0))]
        except ValueError:
            rootID = makeEidEx(1, PrivateFIDs.ROOT if _isPrivate(args) else PublicFIDs.ROOT)
            folders = exmdb.FolderList(client.findFolder(args.folderspec, rootID)).folders
            fids = [folder.folderId for folder in folders]
        if len(fids) == 0:
            cli.print(cli.col("No folders found", "red"))
            return 1
        if len(fids) > 1 and not args.all:
            cli.print(cli.col(f"'{args.folderspec}' is ambiguous. Use -a to delete all or specify the folder ID.", "red"))
            return 2
        for fid in fids:
            try:
                success = client.deleteFolder(fid, args.clear)
                if success:
                    cli.print("Deleted folder 0x{:x}".format(gcToValue(fid)))
                else:
                    cli.print(cli.col("Could not delete folder 0x{:x}".format(gcToValue(fid)), "yellow"))
            except exmdb.ExmdbError:
                cli.print(cli.col("Failed to delete folder 0x{:x}".format(gcToValue(fid)), "yellow"))


def cliExmdbFolderFind(args):
    cli = args._cli
    cli.require("DB")
    from services import Service
    from tools.constants import PublicFIDs, PrivateFIDs
    from tools.rop import makeEidEx
    fid = args.ID or (PrivateFIDs.IPMSUBTREE if _isPrivate(args) else PublicFIDs.IPMSUBTREE)
    fid = makeEidEx(1, fid)
    with Service("exmdb") as exmdb:
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        fuzzyLevel = 0 if args.exact else 65537  # 65537 = SUBSTRING | IGNORECASE
        folders = exmdb.FolderList(client.findFolder(args.name, fid, fuzzyLevel=fuzzyLevel)).folders
        for folder in folders:
            cli.print(_FolderNode(folder).print(cli))
        cli.print(cli.col("({} result{})".format(len(folders), "" if len(folders) == 1 else "s"), attrs=["dark"]))


def cliExmdbFolderList(args):
    cli = args._cli
    cli.require("DB")
    from services import Service
    from tools.constants import PublicFIDs, PrivateFIDs
    from tools.rop import makeEidEx
    fid = args.ID or (PrivateFIDs.IPMSUBTREE if _isPrivate(args) else PublicFIDs.IPMSUBTREE)
    fid = makeEidEx(1, fid)
    with Service("exmdb") as exmdb:
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        root = exmdb.Folder(client.getFolderProperties(0, fid))
        subfolders = exmdb.FolderList(client.listFolders(fid, args.recursive)).folders
        folder = _FolderNode(root, subfolders)
        if args.format in ("csv", "table"):
            from .common import Table
            Table(folder.tabledata(), header=("ID", "parentID", "name"), colsep="," if args.format == "csv" else "  ")\
                .dump(cli, args.format)
        else:
            cli.print(folder.print(cli, args.format))


def _cliExmdbFolderPermissionPrint(cli, permission):
    permstring = "all" if permission & _permsAll == _permsAll else\
                 "none" if permission == 0 else\
                 ",".join(name for name, val in _perms.items() if permission & val)
    return "0x{:04x} ({})".format(permission, cli.col(permstring, attrs=["dark"]))


def cliExmdbFolderPermissionsModify(args):
    cli = args._cli
    cli.require("DB")
    from .common import Table
    from functools import reduce
    from orm.users import Users
    from services import Service
    from tools.rop import makeEidEx
    if not args.revoke:
        if Users.query.filter(Users.username == args.username).count() == 0:
            cli.print(cli.col("Target user '{}' does not exist".format(args.username), "yellow" if args.force else "red"))
            if not args.force:
                return 100
    fid = makeEidEx(1, args.ID)
    perms = reduce(lambda x, y: x | y, args.permission, 0) or _permsAll
    with Service("exmdb") as exmdb:
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        fids = (fid,)
        folders = [exmdb.Folder(client.getFolderProperties(0, fid))]
        if args.recursive:
            folders += exmdb.FolderList(client.listFolders(fid, True)).folders
            fids += tuple(folder.folderId for folder in folders)
        mode = client.REMOVE if args.revoke else client.ADD
        perms = [client.setFolderMember(fid, args.username, perms, mode) for fid in fids]
        cli.print("New permissions for user '{}':".format(cli.col(args.username, attrs=["bold"])))
        Table([(_FolderNode(folder).print(cli), _cliExmdbFolderPermissionPrint(cli, perm))
               for folder, perm in zip(folders, perms)]).print(cli)


def cliExmdbFolderPermissionsShow(args):
    cli = args._cli
    cli.require("DB")
    from .common import Table
    from services import Service
    from tools.rop import makeEidEx
    fid = makeEidEx(1, args.ID)
    with Service("exmdb") as exmdb:
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        members = exmdb.FolderMemberList(client.getFolderMemberList(fid)).members
        Table([(member.mail, _cliExmdbFolderPermissionPrint(cli, member.rights))
               for member in members if not member.special and (not args.username or member.mail in args.username)],
              empty=cli.col("(no entries)", attrs=["dark"])).print(cli)


def cliExmdbStoreGetDelete(args):
    def printSize(value):
        suffix = ("B", "kiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
        index = 0
        prec = 0
        while value > 1000 and index < len(suffix)-1:
            value /= 1024
            index += 1
            prec = 0 if value >= 100 else 1 if value >= 10 else 2 if value >= 1 else 3
        return "{:.{}f} {}".format(value, prec, suffix[index])

    def printVal(pv):
        from datetime import datetime
        from tools.rop import nxTime
        if pv.type == PropTypes.BINARY:
            return cli.col("[{} byte{}]".format(len(pv.val), "" if len(pv.val) == 1 else "s"), attrs=["dark"]), ""
        elif pv.type == PropTypes.FILETIME:
            timestring = datetime.fromtimestamp(nxTime(pv.val)).strftime("%Y-%m-%d %H:%M:%S")
            return pv.val, cli.col(timestring, attrs=["dark"])
        elif pv.type == PropTypes.BINARY_ARRAY:
            return cli.col("[{} value{}]".format(len(pv.val), "" if len(pv.val) == 1 else "s"), attrs=["dark"]), ""
        elif PropTypes.ismv(pv.type):
            return "["+", ".join(repr(val) for val in pv.val)+"]", ""
        return pv.val, cli.col(printSize(pv.val*PropTags.sizeFactor.get(pv.tag, 1)), attrs=["dark"])\
            if pv.tag in PropTags.sizeTags else ""

    cli = args._cli
    cli.require("DB")
    from .common import Table
    from tools.constants import PropTags, PropTypes
    from services import Service
    tags = [PropTags.deriveTag(tag) for tag in args.propspec]
    with Service("exmdb") as exmdb:
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        if not tags:
            tags = client.getAllStoreProperties()
        if args.command == "delete":
            client.removeStoreProperties(tags)
            return
        props = client.getStoreProperties(0, tags)
        Table([(PropTags.lookup(prop.tag, hex(prop.tag)).lower(), *printVal(prop)) for prop in props],
              empty=cli.col("(No properties)", attrs=["dark"])).print(cli)


def cliExmdbStoreSet(args):
    cli = args._cli
    cli.require("DB")
    from tools.constants import PropTags
    from services import Service
    with Service("exmdb") as exmdb:
        props = [exmdb.TaggedPropval(*PropTags.normalize(*prop.split("=", 1))) for prop in args.propspec if "=" in prop]
        ret, client = _getClient(args, exmdb)
        if ret:
            return ret
        client.setStoreProperties(0, props)


def _cliTargetCompleter(prefix, **kwargs):
    from orm.users import Users
    from orm.domains import Domains
    return [user.username for user in Users.query.filter(Users.username.ilike(prefix+"%"))
                                                 .with_entities(Users.username)] +\
           [domain.domainname for domain in Domains.query.filter(Domains.domainname.ilike(prefix+"%"))
                                                         .with_entities(Domains.domainname)]


def _setupCliExmdb(subp: ArgumentParser):
    class PermChoices:
        _permChoices = ("all",)+tuple(_perms.keys())

        def __contains__(self, value):
            return True

        def __getitem__(self, i):
            return self._permChoices[i]

        def __len__(self):
            return len(self._permChoices)

    def xint(x):
        return int(x, 0)

    def perm(x):
        return _permsAll if x.lower() == "all" else _perms.get(x.lower()) or int(x, 0)

    subp.add_argument("target", help="User or domain name").completer = _cliTargetCompleter
    sub = subp.add_subparsers()

    folder = sub.add_parser("folder", help="Access folders")
    Cli.parser_stub(folder)
    foldersub = folder.add_subparsers()
    create = foldersub.add_parser("create", help="Create folder")
    create.set_defaults(_handle=cliExmdbFolderCreate)
    create.add_argument("--comment", default="", help="Folder comment")
    create.add_argument("-t", "--type", default="IPF.Note", help="Folder type (default: IPF.Note)")
    create.add_argument("name", help="Name of the folder to create")
    create.add_argument("ID", nargs="?", type=xint, default=0, help="Parent folder ID")
    delete = foldersub.add_parser("delete", help="Delete folder")
    delete.set_defaults(_handle=cliExmdbFolderDelete)
    delete.add_argument("-a", "--all", action="store_true", help="Delete all matching folders")
    delete.add_argument("--clear", action="store_true", help="Empty folder before deleting")
    delete.add_argument("folderspec", help="ID or name of folder")
    find = foldersub.add_parser("find", help="Find folder by name")
    find.set_defaults(_handle=cliExmdbFolderFind)
    find.add_argument("-x", "--exact", action="store_true", help="Only report exact matches instead of substring matches")
    find.add_argument("name", help="Name of the folder to find")
    find.add_argument("ID", nargs="?", type=xint, help="ID of the folder to search in")
    grant = foldersub.add_parser("grant", help="Grant permissions to user")
    grant.set_defaults(_handle=cliExmdbFolderPermissionsModify, revoke=False)
    grant.add_argument("ID", type=xint, help="Folder ID")
    grant.add_argument("username", help="E-Mail address of the user to grant permissions to").completer = _cliTargetCompleter
    grant.add_argument("permission", nargs="+", type=perm, choices=PermChoices(), help="Permission name or value",
                       metavar="permission")
    grant.add_argument("-f", "--force", action="store_true", help="Add permissions even if user does not exist")
    grant.add_argument("-r", "--recursive", action="store_true", help="Apply to subfolders recursively")
    list = foldersub.add_parser("list", help="List subfolders")
    list.set_defaults(_handle=cliExmdbFolderList)
    list.add_argument("ID", nargs="?", type=xint, default=0, help="Folder ID")
    list.add_argument("--format", nargs="?", default="pretty", help="Output format", metavar="FORMAT",
                      choices=("csv", "json-flat", "json-tree", "pretty", "table"))
    list.add_argument("-r", "--recursive", action="store_true", help="Recursively list subfolders")
    permissions = foldersub.add_parser("permissions")
    permissions.set_defaults(_handle=cliExmdbFolderPermissionsShow)
    permissions.add_argument("ID", type=xint, help="Folder ID")
    permissions.add_argument("username", nargs="*", help="E-Mail address of the user to show permissions of")\
               .completer = _cliTargetCompleter
    revoke = foldersub.add_parser("revoke", help="Grant permissions to user")
    revoke.set_defaults(_handle=cliExmdbFolderPermissionsModify, revoke=True)
    revoke.add_argument("ID", type=xint, help="Folder ID")
    revoke.add_argument("username", help="E-Mail address of the user to revoke permissions from")\
          .completer = _cliTargetCompleter
    revoke.add_argument("permission", nargs="*", type=perm, choices=PermChoices(), help="Permission name or value",
                        metavar="permission")
    revoke.add_argument("-r", "--recursive", action="store_true", help="Apply to subfolders recursively")

    store = sub.add_parser("store", help="Access store properties")
    Cli.parser_stub(store)
    storesub = store.add_subparsers()
    get = storesub.add_parser("get", help="Query store properties")
    get.set_defaults(_handle=cliExmdbStoreGetDelete, command="get")
    get.add_argument("propspec", nargs="*", help="Properties to query").completer = proptagCompleter
    delete = storesub.add_parser("delete", help="Delete store properties")
    delete.set_defaults(_handle=cliExmdbStoreGetDelete, command="delete")
    delete.add_argument("propspec", nargs="+", help="Properties to delete").completer = proptagCompleter
    set = storesub.add_parser("set", help="Set store properties")
    set.set_defaults(_handle=cliExmdbStoreSet)
    set.add_argument("propspec", nargs="+", help="Properties to query", metavar="propspec=value")\
       .completer = lambda *args, **kwargs: proptagCompleter(*args, **kwargs, addSuffix="=")


@Cli.command("exmdb", _setupCliExmdb, help="Private/public store management")
def cliExmdbStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from flask import Flask, jsonify, request, make_response
from functools import wraps

from orm import DB
from services import Service
from tools.config import Config

from . import apiSpec


class OpenApiCompat:
    def __init__(self, apiSpec):
        import openapi_core
        self.version = [int(part) for part in openapi_core.__version__.split(".")]
        if self.version < [0, 13, 0]:
            from openapi_core.wrappers.flask import FlaskOpenAPIRequest, FlaskOpenAPIResponse
        else:
            from openapi_core.contrib.flask import FlaskOpenAPIRequest, FlaskOpenAPIResponse
        if self.version < [0, 15, 0]:
            from openapi_core import create_spec
        else:
            from openapi_core.spec.shortcuts import create_spec
        self.spec = create_spec(apiSpec)
        if self.version < [0, 15, 0]:
            from openapi_core.shortcuts import RequestValidator, ResponseValidator
            self.requestValidator = RequestValidator(self.spec)
            self.responseValidator = ResponseValidator(self.spec)
            self.validateRequest = lambda request: self.requestValidator.validate(FlaskOpenAPIRequest(request))
            self.validateResponse = lambda request, response: \
                self.responseValidator.validate(FlaskOpenAPIRequest(request), FlaskOpenAPIResponse(response)).errors
        else:
            self.FlaskOpenAPIRequest, self.FlaskOpenAPIResponse = FlaskOpenAPIRequest, FlaskOpenAPIResponse
            self.validateRequest = self._validateRequest_15_0
            self.validateResponse = self._validateResponse_15_0

    @staticmethod
    def _suppressError(exc):
        def matchBuggedError(err):
            return isinstance(err, ValidationError) and err.message == "None for not nullable" and "$ref" in err.schema
        from openapi_core.unmarshalling.schemas.exceptions import InvalidSchemaValue
        from jsonschema.exceptions import ValidationError
        if not isinstance(exc, InvalidSchemaValue):
            return False
        return all(matchBuggedError(err) for err in exc.schema_errors)

    def _validateRequest_15_0(self, request):
        from openapi_core.validation.request import openapi_request_validator as reqval
        result = reqval.validate(self.spec, self.FlaskOpenAPIRequest(request))
        return result

    def _validateResponse_15_0(self, request, response):
        from openapi_core.validation.response import openapi_response_validator as resval
        result = resval.validate(self.spec, self.FlaskOpenAPIRequest(request), self.FlaskOpenAPIResponse(response))
        return [error for error in result.errors if not self._suppressError(error)]


if "servers" in Config["openapi"]:
    apiSpec["servers"] += Config["openapi"]["servers"]

validator = OpenApiCompat(apiSpec)

API = Flask("grommunio Admin API")  # Core API object
API.config["JSON_SORT_KEYS"] = False  # Do not sort response fields. Crashes when returning lists...
if DB is not None:
    DB.enableFlask(API)

if not Config["openapi"]["validateRequest"]:
    API.logger.warning("Request validation is disabled!")
if not Config["openapi"]["validateResponse"]:
    API.logger.warning("Response validation is disabled!")


def validateRequest(flask_request):
    """Validate the request

    Parameters
    ----------
    flask_request: flask.request
        The request sent by flask

    Returns
    -------
    Boolean
        True if the request is valid, False otherwise
    string
        Error message if validation failed, None otherwise"""
    result = validator.validateRequest(flask_request)
    if result.errors:
        return False, jsonify(message="Bad Request", errors=[type(error).__name__ for error in result.errors]), result.errors
    return True, None, None


def reloadORM():
    """Reload all active orm modules."""
    import importlib
    import sys
    API.logger.warn("Database schema version updated detected - reloading ORM")
    DB.initVersion()
    for name, module in [(name, module) for name, module in sys.modules.items() if name.startswith("orm.")]:
        importlib.reload(module)


def secure(requireDB=False, requireAuth=True, authLevel="basic", service=None, validateCSRF=None):
    """Decorator securing API functions.

       Arguments:
           - requireDB (boolean or int)
               Whether the database is needed for the call. If set to True and the database is not configured,
               and error message is returned without invoking the endpoint. If given as an integer, marks the minimum required
               schema version.
           - requireAuth (boolean or "optional")
               Whether authentication is required to use this endpoint. When set to False, no login context is created
               and user information is not available, even if logged in.
           - authLevel ("basic" or "user")
               Create login context with user object ("user") or only with information from token ("basic").
               User information can be loaded later if necessary.
           - service (string)
               Execute this endpoint in a service context. The service object is passed to the endpoint function
               as the last (unnamed) parameter.
           - validateCSRF (bool or None)
               Validate CSRF token. None will enable validation for non-GET methods.

       Automatically validates the request using the OpenAPI specification and returns a HTTP 400 to the client if validation
       fails. Also validates the response generated by the endpoint and returns a HTTP 500 on error. This behavior can be
       deactivated in the configuration.

       If an exception is raised during execution, a HTTP 500 message is returned to the client and a short description of the
       error is sent in the 'error' field of the response.
       """
    from .security import getSecurityContext

    def inner(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            def call():
                if service:
                    with Service(service) as srv:
                        ret = func(*args, srv, **kwargs)
                else:
                    ret = func(*args, **kwargs)
                response = make_response(ret)
                try:
                    result = validator.validateResponse(request, response)
                except AttributeError:
                    result = None
                if result:
                    if Config["openapi"]["validateResponse"]:
                        API.logger.error("Response validation failed: "+str(result))
                        return jsonify(message="The server generated an invalid response."), 500
                    else:
                        API.logger.warn("Response validation failed: "+str(result))
                return ret

            if requireAuth:
                checkCSRF = False if Config["security"].get("disableCSRF") else validateCSRF
                error = getSecurityContext(authLevel, checkCSRF)
                if error is not None and requireAuth != "optional":
                    return jsonify(message="Access denied", error=error), 401
            valid, message, errors = validateRequest(request)
            if not valid:
                if Config["openapi"]["validateRequest"]:
                    API.logger.info("Request validation failed: {}".format(errors))
                    return message, 400
                else:
                    API.logger.warn("Request validation failed: {}".format(errors))

            if requireDB or requireAuth:
                if DB is None:
                    return jsonify(message="Database not available."), 503
                if DB.requireReload():
                    reloadORM()
                if isinstance(requireDB, int) and not isinstance(requireDB, bool) and DB.version < requireDB:
                    return jsonify(message="Database schema version too old. Please update to at least n{}."
                                   .format(requireDB)), 500
            return call()
        return wrapper
    return inner


@API.after_request
def noCache(response):
    """Add no-cache headers to the response"""
    response.cache_control.no_cache = True
    response.cache_control.no_store = True
    response.cache_control.max_age = 1
    return response


from . import errors as _



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020-2021 grommunio GmbH


BaseRoute = "/api/v1"  # Common prefix for all endpoints

apiSpec = None  # API specification
apiVersion = None  # API specification version. Extracted from the OpenAPI document.
backendVersion = "1.11.19"  # Backend version number


def _loadOpenApiSpec():
    global apiVersion, apiSpec
    try:
        import json
        with open("res/openapi.json", "r", encoding="utf-8") as file:
            apiSpec = json.load(file)
    except FileNotFoundError:
        import yaml
        with open("res/openapi.yaml", "r", encoding="utf-8") as file:
            apiSpec = yaml.load(file, Loader=yaml.SafeLoader)
    apiVersion = apiSpec["info"]["version"]


_loadOpenApiSpec()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from random import SystemRandom
from .constants import SERIAL_ENDIAN

class GUID:
    rndEng = SystemRandom()

    def __init__(self):
        pass

    @staticmethod
    def fromDomainID(domainID : int):
        guid = GUID()
        guid.timeLow: int = domainID
        guid.timeMid: int = 0xafb
        guid.timeHiAndVersion: int = 0x7df6
        guid.clockSeq: tuple = (0x91, 0x92)
        guid.node: tuple = (0x49, 0x88, 0x6a, 0xa7, 0x38, 0xce)
        return guid

    @staticmethod
    def random():
        guid = GUID()
        guid.timeLow: int = GUID.rndEng.getrandbits(32)
        guid.timeMid: int = GUID.rndEng.getrandbits(16)
        guid.timeHiAndVersion: int = GUID.rndEng.getrandbits(16) & 0x0FFF | 0x4000
        guid.clockSeq: tuple = (GUID.rndEng.getrandbits(8) & 0x3f | 0x80, GUID.rndEng.getrandbits(8))
        guid.node: tuple = tuple(GUID.rndEng.getrandbits(8) for _ in range(6))
        return guid

    def serialize(self) -> bytes:
        return self.timeLow.to_bytes(4, SERIAL_ENDIAN)+\
               self.timeMid.to_bytes(2, SERIAL_ENDIAN)+\
               self.timeLow.to_bytes(2, SERIAL_ENDIAN)+\
               bytes(self.clockSeq)+bytes(self.node)

    def __str__(self):
        return "{:08x}-{:04x}-{:04x}-{:02x}{:02x}-{:02x}{:02x}{:02x}{:02x}{:02x}{:02x}".format(self.timeLow,
                                                                                               self.timeMid,
                                                                                               self.timeHiAndVersion,
                                                                                               *self.clockSeq,
                                                                                               *self.node)

class XID:
    def __init__(self):
        pass

    @staticmethod
    def fromDomainID(domainID: int, changeID: int):
        xid = XID()
        xid.guid = GUID.fromDomainID(domainID)
        xid.localID = tuple(i for i in changeID.to_bytes(6, "big"))
        return xid

    def serialize(self) -> bytes:
        return self.guid.serialize()+bytes(self.localID);



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from flask import jsonify, request

import api
from api.core import API, secure
from api.security import checkPermissions

from tools.permissions import Permissions, SystemAdminPermission, SystemAdminROPermission
from .. import defaultListHandler, defaultObjectHandler, userQuery


@API.route(api.BaseRoute+"/system/users", methods=["GET"])
@secure(requireDB=True)
def userListEndpointUnrestricted():
    checkPermissions(SystemAdminROPermission())
    return userQuery()


@API.route(api.BaseRoute+"/system/roles/permissions", methods=["GET"])
@secure()
def getAdminPermissions():
    checkPermissions(SystemAdminROPermission())
    return jsonify(data=Permissions.knownPermissions())


@API.route(api.BaseRoute+"/system/roles", methods=["GET", "POST"])
@secure(requireDB=True, authLevel="user")
def adminRolesListEndpoint():
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    from orm.roles import AdminRoles
    return defaultListHandler(AdminRoles)


@API.route(api.BaseRoute+"/system/roles/<int:ID>", methods=["GET", "PATCH", "DELETE"])
@secure(requireDB=True, authLevel="user")
def adminRolesObjectEndpoint(ID):
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    from orm.roles import AdminRoles, AdminUserRoleRelation
    if request.method == "DELETE" and AdminUserRoleRelation.query.filter(AdminUserRoleRelation.roleID == ID).count() > 0:
        return jsonify(message="Cannot delete role that is still in use"), 400
    return defaultObjectHandler(AdminRoles, ID, "Role")



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH
"""Create fetchmail configuration from database"""

from . import Cli, InvalidUseError
from .common import userspecAutocomp, userCandidates
from argparse import ArgumentParser


_globalConf = "# Global settings\n" \
              "set postmaster \"postmaster\"\n" \
              "set nobouncemail\n"\
              "set no spambounce\n"\
              "set properties \"\"\n"\
              "\n# Accounts\n"

_auths = ("password", "kerberos_v5", "kerberos", "kerberos_v4", "gssapi", "cram-md5", "otp", "ntlm", "msn", "ssh", "any")
_proto = ("POP3", "IMAP", "POP2", "ETRN", "AUTO")


def _sanitizeData(data):
    _cliargs = {"_handle", "_cli", "userspec", "mbspec"}
    return {key: value for key, value in data.items() if key not in _cliargs and value is not None}


def _dumpFml(cli, fml, passwd=False):
    cli.print(cli.col("{} ({}):".format(fml.mailbox, fml.ID), attrs=["bold"]))
    cli.print("  user: {} ({})".format(fml.user.username, fml.user.ID))
    cli.print("  active: "+(cli.col("yes", "green") if fml.active == 1 else cli.col("no", "red")))
    cli.print("  changed: "+fml.date.ctime())
    cli.print("  srcUser: "+fml.srcUser)
    if passwd:
        cli.print("  srcPassword: "+fml.srcPassword)
    for attr in ("srcFolder", "srcServer", "srcAuth", "protocol"):
        val = getattr(fml, attr)
        cli.print("  {}: {}".format(attr, val if val else cli.col("(unset)", attrs=["dark"])))
    for attr in ("fetchall", "keep", "useSSL"):
        cli.print("  {}: {}".format(attr, "yes" if  getattr(fml, attr) else "no"))
    if fml.useSSL == 1:
        cli.print("  sslCheckCert: "+("yes" if fml.sslCertCheck == 1 else "no"))
        cli.print("  sslFingerprint: "+(fml.sslFingerprint or cli.col("(unset)", attrs=["dark"])))
        cli.print("  sslCertPath: "+(fml.sslCertPath or cli.col("(unset)", attrs=["dark"])))
    cli.print("  extraOptions: "+(fml.extraOptions or cli.col("(none)", attrs=["dark"])))


def cliCreateFetchmail(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from orm.users import Fetchmail
    users = userCandidates(args.userspec).all()
    if len(users) == 0:
        cli.print(cli.col("User not found.", "yellow"))
        return 1
    if len(users) > 1:
        cli.print(cli.col("'{}' is ambiguous.".format(args.userspec), "yellow"))
        return 2
    user = users[0]
    if args.mailbox is None:
        args.mailbox = user.username
    data = (_sanitizeData(args.__dict__))
    err = Fetchmail.checkCreateParams(data)
    if err is not None:
        cli.print(cli.col("Cannot create fetchmail entry: "+err, "red"))
        return 3
    try:
        fml = Fetchmail(data, user)
        DB.session.add(fml)
        DB.session.commit()
        _dumpFml(cli, fml)
    except BaseException as err:
        cli.print(cli.col("Cannot create fetchmail entry: "+" - ".join(str(arg) for arg in err.args), "red"))
        DB.session.rollback()
        return 4


def cliDeleteFetchmail(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from orm.users import Fetchmail
    fmls = Fetchmail.query.filter(Fetchmail.ID == args.mbspec if args.mbspec.isdigit() else
                                  Fetchmail.mailbox.ilike(args.mbspec+"%")).all()
    if len(fmls) == 0:
        cli.print(cli.col("Fetchmail entry not found.", "yellow"))
        return 1
    if len(fmls) > 1 and not args.yes:
        prompt = "Delete multiple entries:\n  "+"\n  ".join("{} ({})".format(fml.mailbox, fml.ID) for fml in fmls)+"\n[y/N]: "
        if cli.confirm(prompt) != Cli.SUCCESS:
            return 2
    try:
        for fml in fmls:
            DB.session.delete(fml)
        DB.session.commit()
    except BaseException as err:
        cli.print(cli.col("Deletion failed: "+" - ".join(str(arg) for arg in err.args), "red"))
        DB.session.rollback()
        return 4
    cli.print("{} entr{} deleted.\n".format(len(fmls), "y" if len(fmls) == 1 else "ies")+
              cli.col("Entry deletion will not be handled automatically, consider running ", "yellow")+
              cli.col("write-rc -f", "yellow", attrs=["bold", "dark"])+
              cli.col(" to manually update the configuration.", "yellow"))


def cliListFetchmail(args):
    cli = args._cli
    cli.require("DB")
    from orm.users import Fetchmail
    query = Fetchmail.optimized_query(1)
    if "filter" in args and args.filter is not None:
        query = Fetchmail.autofilter(query, {f.split("=", 1)[0]: f.split("=", 1)[1] for f in args.filter if "=" in f})
    if "sort" not in args or args.sort is None:
        args.sort = "active,desc"
    query = Fetchmail.autosort(query, args.sort)
    if args.mbspec is not None:
        query = query.filter(Fetchmail.ID == args.mbspec if args.mbspec.isdigit() else Fetchmail.mailbox.ilike(args.mbspec+"%"))
    entries = query.all()
    if len(entries) == 0:
        cli.print(cli.col("No results", "yellow"))
    for entry in entries:
        cli.print("{} ({}) from '{}@{}/{}', {}, updated {}".format(
                cli.col(entry.mailbox, attrs=["bold"]),
                cli.col(entry.ID, attrs=["bold"]),
                entry.srcUser,
                entry.srcServer,
                entry.srcFolder,
                cli.col("active", "green") if entry.active == 1 else cli.col("inactive", "red"),
                entry.date.ctime()))


def cliModifyFetchmail(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from orm.users import Fetchmail
    fmls = Fetchmail.query.filter(Fetchmail.ID == args.mbspec if args.mbspec.isdigit() else
                                  Fetchmail.mailbox.ilike(args.mbspec+"%")).all()
    if len(fmls) == 0:
        cli.print(cli.col("Fetchmail entry not found.", "yellow"))
        return 1
    if len(fmls) > 1:
        cli.print(cli.col("'{}' is ambiguous.".format(args.mbspec), "yellow"))
        return 2
    fml = fmls[0]
    data = _sanitizeData(args.__dict__)
    try:
        fml.fromdict(data)
        DB.session.commit()
        _dumpFml(cli, fml)
    except BaseException as err:
        cli.print(cli.col("Cannot create fetchmail entry: "+" - ".join(str(arg) for arg in err.args), "red"))
        DB.session.rollback()
        return 3


def cliPrintFetchmail(args):
    cli = args._cli
    cli.require("DB")
    from orm.users import Fetchmail
    fmls = Fetchmail.query.filter(Fetchmail.ID == args.mbspec if args.mbspec.isdigit() else
                                  Fetchmail.mailbox.ilike(args.mbspec+"%")).all()
    if len(fmls) == 0:
        cli.print(cli.col("No fetchmail entries found.", "yellow"))
    for fml in fmls:
        cli.print(("{} ({}): ".format(fml.mailbox, fml.ID) if not args.quiet else "")+cli.col(str(fml), attrs=["bold"]), end="")


def cliShowFetchmail(args):
    cli = args._cli
    cli.require("DB")
    from orm.users import Fetchmail
    from sqlalchemy.orm import joinedload
    fmls = Fetchmail.query.filter(Fetchmail.ID == args.mbspec if args.mbspec.isdigit() else
                                  Fetchmail.mailbox.ilike(args.mbspec+"%")).options(joinedload(Fetchmail.user)).all()
    if len(fmls) == 0:
        cli.print(cli.col("No fetchmail entries found.", "yellow"))
    for fml in fmls:
        _dumpFml(cli, fml, args.password)


def cliWriteFetchmailrc(args):
    cli = args._cli
    def write(data):
        file.write(data)
        if args.print:
            cli.print(cli.col(data, attrs=["bold"]), end="")

    def vprint(*vargs, **vkwargs):
        if args.verbose:
            cli.print(*vargs, **vkwargs)

    cli.require("DB")
    from datetime import datetime, timedelta
    from orm.users import Fetchmail
    if not args.force:
        if args.time == "auto":
            if cli.fs is not None:
                vprint("Filesystem is emulated - update autdetection skipped")
                mtime = None
            else:
                try:
                    import os
                    mtime = datetime.utcfromtimestamp(os.path.getmtime(args.out_file))
                    vprint("Last fetchmailrc modification was on "+mtime.ctime())
                except:
                    mtime = None
                    vprint("No fetchmailrc found, creating")
        else:
            try:
                mtime = datetime.now()-timedelta(minutes=int(args.time))
                vprint("Manually set last update to "+mtime.ctime())
            except:
                vprint(cli.col("Invalid time specification. Must be integer or 'auto'.", "red"))
                return 3
        if mtime is not None and Fetchmail.query.filter(Fetchmail.date > mtime).count() == 0:
            cli.print(cli.col("No new accounts created since {}. Use -f to force creation.".format(mtime.ctime()), "yellow"))
            return 0
    try:
        vprint("Writing output to "+args.out_file)
        with cli.open(args.out_file, "w") as file:
            write(_globalConf)
            for record in Fetchmail.query.filter(Fetchmail.active == 1).all():
                write(str(record))
    except OSError as err:
        cli.print(cli.col("Could not write to file: {} - {}".format(err.errno, err.strerror), "red"))
        return 2
    except Exception as err:
        cli.print(cli.col("Could not write to file: "+"-".join(str(arg) for arg in err.args)))
        return 3
    if cli.fs is None:
        from tools.misc import setDirectoryOwner, setDirectoryPermission
        try:
            setDirectoryPermission(args.out_file, 0o600)
        except Exception as err:
            cli.print(cli.col("Failed to set file permission: {} - {}"\
                              .format(type(err).__name__, " - ".join(str(arg) for arg in err.args)), "yellow"))
        try:
            setDirectoryOwner(args.out_file, "fetchmail")
        except Exception as err:
            cli.print(cli.col("Failed to set file owner: {} - {}"\
                              .format(type(err).__name__, " - ".join(str(arg) for arg in err.args)), "yellow"))


def _fmlAutocomp(prefix, **kwargs):
    try:
        from orm.users import Fetchmail
        return (fml.mailbox
                for fml in Fetchmail.query.with_entities(Fetchmail.mailbox).filter(Fetchmail.mailbox.ilike(prefix+"%")))
    except:
        return ()

def _setupCliFetchmailParser(subp: ArgumentParser):
    def getBool(val):
        if not val.isdigit() and val not in ("yes", "no"):
            return val
        return bool(int(val)) if val.isdigit() else val == "yes"

    def addParameters(parser: ArgumentParser, init: bool=False):
        idef = lambda x: x if init else None
        bvals = (0, 1, "yes", "no")
        parser.add_argument("--active", default=idef(1), type=getBool, choices=bvals, help="Whether to activate the entry")
        parser.add_argument("--extraOptions", help="Additional fetchmail options")
        parser.add_argument("--fetchall", default=idef(0), type=getBool, choices=bvals, help="Also fetch seen mails")
        parser.add_argument("--keep", default=idef(1), type=getBool, choices=bvals, help="Keep fetched mails on the source server")
        parser.add_argument("--protocol", default=idef("IMAP"), choices=_proto, help="Protocol to use")
        parser.add_argument("--srcAuth", default=idef("password"), choices=_auths, help="Source server authentication")
        parser.add_argument("--srcFolder", help="Source folder")
        parser.add_argument("--srcPassword", required=init, help="Source user password")
        parser.add_argument("--srcServer", required=init, help="Source server address")
        parser.add_argument("--srcUser", required=init, help="Source user")
        parser.add_argument("--sslCertCheck", default=idef(0), type=getBool, choices=bvals, help="Force SSL certificate check")
        parser.add_argument("--sslCertPath", help="Path to certificate directory or empty for system default")
        parser.add_argument("--useSSL", default=idef(1), type=getBool, choices=bvals, help="Enable SSL")

    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    create = sub.add_parser("create", help="Create new fetchmail entry")
    create.add_argument("userspec", help="Target user").completer = userspecAutocomp
    create.add_argument("mailbox", nargs="?", help="Local mailbox address. Defaults to target user.")
    create.set_defaults(_handle=cliCreateFetchmail)
    addParameters(create, True)
    delete = sub.add_parser("delete", help="Delete fetchmail entry")
    delete.set_defaults(_handle=cliDeleteFetchmail)
    delete.add_argument("mbspec", nargs="?", help="Name of the mailbox or ID of the entry").completer = _fmlAutocomp
    delete.add_argument("-y", "--yes", action="store_true", help="Delete multiple entries without confirmation")
    list = sub.add_parser("list", help="List fetchmail entries")
    list.set_defaults(_handle=cliListFetchmail)
    list.add_argument("mbspec", nargs="?", help="Mailbox prefix or ID").completer = _fmlAutocomp
    list.add_argument("-f", "--filter", nargs="*", help="Filter by attribute, e.g. -f ID=42")
    list.add_argument("-s", "--sort", nargs="*", help="Sort by attribute, e.g. -s mailbox,desc")
    modify = sub.add_parser("modify", help="Modify fetchmail entry")
    modify.set_defaults(_handle=cliModifyFetchmail)
    modify.add_argument("mbspec", help="Name of the mailbox or ID of the entry").completer = _fmlAutocomp
    modify.add_argument("--mailbox", help="Local mailbox address")
    addParameters(modify)
    print = sub.add_parser("print", help="Print configuration line generated by fetchmail entry")
    print.set_defaults(_handle=cliPrintFetchmail)
    print.add_argument("mbspec", help="Name of the mailbox or ID of the entry").completer = _fmlAutocomp
    print.add_argument("-q", "--quiet", action="store_true", help="Do not print additional info")
    show = sub.add_parser("show", help="Show detailed information about fetchmail entry")
    show.set_defaults(_handle=cliShowFetchmail)
    show.add_argument("mbspec", help="Name of the mailbox or ID of the entry").completer = _fmlAutocomp
    show.add_argument("--password", action="store_true", help="Print source user password")
    writerc = sub.add_parser("write-rc", help="Write fetchmail configuration file")
    writerc.set_defaults(_handle=cliWriteFetchmailrc)
    writerc.add_argument("-f", "--force", action="store_true", help="Update even if no new accounts were created")
    writerc.add_argument("-o", "--out-file", metavar="FILE", default="/etc/fetchmailrc", help="Override output file")
    writerc.add_argument("-p", "--print", action="store_true", help="Print output")
    writerc.add_argument("-t", "--time", default="auto", help="Time since last update (minutes), or 'auto'")
    writerc.add_argument("-v", "--verbose", action="store_true", help="Be more verbose")


@Cli.command("fetchmail", _setupCliFetchmailParser, help="Fetchmail management")
def cliFetchmailStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

import time

from datetime import datetime
from flask import jsonify, request

import api
from api.core import API, secure
from api.errors import InsufficientPermissions
from api.security import checkPermissions

from services import Service

from tools.config import Config
from tools.permissions import DomainAdminROPermission, SystemAdminPermission


def checkAccess(permission):
    """Check whether the request comes from an permitted source or user has sufficient permissions."""
    request.remote_addr in Config["sync"]["policyHosts"] or checkPermissions(permission)


@API.route(api.BaseRoute+"/service/syncPolicy/<username>", methods=["GET"])
@secure(requireDB=True, requireAuth="optional")
def getUserSyncPolicy(username):
    checkAccess(DomainAdminROPermission("*"))
    from orm.domains import Domains
    from orm.users import Users
    user = Users.query.filter(Users.username == username).first()
    if user is None:
        return jsonify(data=Config["sync"]["defaultPolicy"])
    checkAccess(DomainAdminROPermission(user.domainID))
    domain = Domains.query.filter(Domains.ID == user.domainID, Domains._syncPolicy != None).first()
    policy = dict(Config["sync"]["defaultPolicy"])
    if domain is not None:
        policy.update(domain.syncPolicy)
    if user.syncPolicy is not None:
        policy.update(user.syncPolicy)
    return jsonify(data=policy)


@API.route(api.BaseRoute+"/service/wipe/<username>", methods=["GET"])
@secure(requireDB=True, requireAuth="optional")
def getWipeStatus(username):
    from orm.users import Users, UserDevices
    user = Users.query.filter(Users.username == username).first()
    if user is None:
        return jsonify(message="User not found"), 404
    checkAccess(DomainAdminROPermission(user.domainID))
    try:
        devices = {device.deviceID: device.status
                   for device in UserDevices.query.filter(UserDevices.userID == user.ID)}
    except Exception:
        devices = {}
    requested = request.args["devices"].split(",") if "devices" in request.args else None
    if requested:
        data = {deviceID: {"status": devices[deviceID]} if deviceID in devices else UserDevices.DEFAULT
                for deviceID in requested}
    else:
        data = {deviceID: {"status": status} for deviceID, status in devices.items()}
    return jsonify(data=data)


@API.route(api.BaseRoute+"/service/wipe/<username>", methods=["POST"])
@secure(requireDB=True, requireAuth="optional", authLevel="user")
def setWipeStatus(username):
    checkAccess(SystemAdminPermission())
    from orm.users import DB, Users, UserDevices, UserDeviceHistory
    if not DB.minVersion(93):
        return jsonify(message="Database schema too old - please update to at least n93"), 503
    user = Users.query.filter(Users.username == username).first()
    if user is None:
        return jsonify(message="User not found"), 404
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing data")
    requested = set(request.args["devices"].split(",")) if "devices" in request.args else None
    devices = {device.deviceID: device for device in UserDevices.query.filter(UserDevices.userID == user.ID)
               if not requested or device.deviceID in requested}
    authenticated = False
    if "password" in data:
        if "user" in request.auth:
            authenticated = request.auth["user"].chkPw(data["password"])
        else:
            authenticated = user.chkPw(data["password"])
    newStatus = data["status"]
    if requested:
        for deviceID in requested:
            if deviceID not in devices:
                if newStatus > 1 and not authenticated:
                    raise InsufficientPermissions()
                devices[deviceID] = UserDevices(dict(userID=user.ID, deviceID=deviceID, status=newStatus))
                DB.session.add(devices[deviceID])
    DB.session.flush()
    timestamp = datetime.fromtimestamp(data["time"]) if "time" in data else datetime.utcnow()
    remote = data.get("remoteIP", request.remote_addr)
    refresh = []
    for device in devices.values():
        if device.status <= 1 and newStatus > 1:
            if not authenticated:
                raise InsufficientPermissions()
            refresh.append(device.deviceID+"|-|"+user.username)
        device.status = newStatus
        DB.session.add(UserDeviceHistory(dict(userDeviceID=device.ID, time=timestamp, remoteIP=remote, status=newStatus)))
    DB.session.commit()
    if refresh:
        with Service("redis") as redis:
            redis.hdel("grommunio-sync:provisioningcache", *refresh)
    return jsonify(message="Success."), 201


@API.route(api.BaseRoute+"/service/userinfo/<username>", methods=["GET"])
@secure(requireDB=True, requireAuth="optional", authLevel="user")
def getUserInfo(username):
    from orm.users import Users
    user = Users.query.filter(Users.username == username).with_entities(Users.domainID, Users.externID).first()
    if user is None:
        return jsonify(message="User not found"), 404
    checkAccess(DomainAdminROPermission(user.domainID))
    return jsonify(ldap=user.externID is not None)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

__all__ = ["domain", "system", "defaults", "misc", "service", "tasq"]

from flask import request, jsonify
from orm import DB
from tools.DataModel import MissingRequiredAttributeError, InvalidAttributeError, MismatchROError
from tools.misc import damerau_levenshtein_distance as dldist
import re

from sqlalchemy.exc import IntegrityError

matchStringRe = re.compile(r"([\w\-]*)")


def defaultListQuery(Model, filters=(), order=None, result="response", automatch=True, autofilter=True, autosort=True,
                     include_count="count", query=None):
    """Process a listing query for specified model.

    Automatically uses 'limit' (50), 'offset' (0) and 'level' (1) parameters from the request.

    The return value can be influenced by `result`: `list` will return a list ob objects, while the default `response`
    will return the complete JSON encoded flask response.

    If `automatch` is enabled, the results are filtered by prefix-matching each word against the configured columns. If no
    other sorting is active (`order` is None and no "sort" query parameter is given), the results are ranked by the
    Damerau-Levenshtein distance to the search term. Note that ranking is done after the query and a low `limit` parameter
    may prevent a good match from being selected at all.

    Parameters
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to perform the query on
    filters : iterable, optional
        A list of filter expressions to apply to the query. The default is ().
    order : list or Column, optional
        Column(s) to use in an order_by expression. The default is None.
    result : str, optional
        Return type.
    automatch : str, optional
        Name of the column to match against.
    autofilter : list of str, optional
        Whether to apply autofiltering. See DataModel.autofilter for more information. Default is True.
    autosort: boolean, optional
        Whether to apply autosorting. See DataModel.autosort for more information. Default is True.
    include_count: str, optional
        Name of the property containing the total number of results (regardless of limit) or None to disable.
        Default is "count".
    query: BaseQuery, optional
        Specify a base query to build upon. Default is None.
    Returns
    -------
    Response
        Flask response containing the list data.
    """
    limit = request.args.get("limit", "50")
    if len(limit) == 0:
        limit = None
    offset = request.args.get("offset", "0")
    if len(offset) == 0:
        offset = None
    verbosity = int(request.args.get("level", 1))
    query = (Model.optimized_query(verbosity) if query is None else Model.optimize_query(query, verbosity)).filter(*filters)
    if autosort:
        query = Model.autosort(query, request.args.getlist("sort"))
    if order is not None:
        query = query.order_by(*(order if type(order) in (list, tuple) else (order,)))
    if autofilter:
        query = Model.autofilter(query, request.args)
    if automatch and "match" in request.args:
        matchStr = request.args["match"].lower()
        fields = set(request.args["matchFields"].split(",")) if "matchFields" in request.args else None
        query = Model.automatch(query, request.args["match"], fields)
    count = query.count() if include_count else None
    if result == "query":
        return query, limit, offset, count
    query = query.limit(limit).offset(offset)
    objects = query.all()
    if order is None and "sort" not in request.args and automatch and "match" in request.args:
        scored = ((min(dldist(str(field).lower(), matchStr) for field in obj.matchvalues(fields) if field is not None), obj)
                  for obj in objects)
        objects = [so[1] for so in sorted(scored, key=lambda entry: entry[0])]
    if result == "list":
        return objects
    data = [obj.todict(verbosity) for obj in objects]
    if result == "data":
        return data
    resp = dict(data=data)
    if include_count:
        resp[include_count] = count
    return jsonify(resp)


def defaultDetailQuery(Model, ID, errName, filters=()):
    """Process a detail query for specified model.

    Automatically uses 'level' (2) parameter from the request.
    Returns a 404 error response if no object with the given ID is found.

    Parameters
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to perform the query on.
    ID : int
        ID of the object.
    errName : str
        Object name to use in error messages.

    Returns
    -------
    Response
        Flask response containing the object data or an error message.
    """
    verbosity = request.args.get("level", 2)
    query = Model.query.filter(Model.ID == ID, *filters)
    query = Model.optimize_query(query, verbosity)
    obj = query.first()
    if obj is None:
        return jsonify(message=errName+" not found"), 404
    return jsonify(obj.todict(verbosity))


def defaultPatch(Model, ID, errName, obj=None, filters=(), result="response"):
    """Process a PATCH query for specified model.

    Performs an autopatch() call on the model.
    Returns a 404 error response if no object with the given ID is found.

    Parameters
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to perform the query on.
    ID : int
        ID of the object.
    errName : str
        Object name to use in error messages.
    obj : SQLAlchemy model instance, optional
        Object to patch (suppresses object retrieval query). The default is None.
    response: str
        Return value. "precommit" returns the patched object before committing the changes. Other values return the response.

    Returns
    -------
    <varies>
        Return value according to `result`
    """
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Could not update: no valid JSON data"), 400
    if obj is None:
        obj = Model.query.filter(Model.ID == ID, *filters).first()
    if obj is None:
        return jsonify(message=errName+" not found"), 404
    try:
        obj.fromdict(data)
    except (InvalidAttributeError, MismatchROError, ValueError) as err:
        DB.session.rollback()
        return jsonify(message=err.args[0]), 400
    if result == "precommit":
        return obj
    try:
        DB.session.commit()
    except IntegrityError as err:
        DB.session.rollback()
        return jsonify(message="Could not update: invalid data", error=err.orig.args[1]), 400
    return jsonify(Model.optimized_query(2).filter(Model.ID == ID).first().fulldesc())


def defaultCreate(Model, result="response"):
    """Create a new object of the specified model.

    Performs a check on the input data, calls Model ctor and tries to insert the object into the database.
    Catches the following errors an returns an appropriate error message with HTTP 400 code:
        - The request does not contain a valid JSON object
        - Parameter check fails
        - Ctor raises a MissingRequiredAttributeError
        - Database commit raises an IntegrityError

    ParametersDefault
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to create a new instance from

    Returns
    -------
    Response
        Flask response containing the new object data or an error message.
    """
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Invalid JSON object"), 400
    error = Model.checkCreateParams(data)
    if error is not None:
        return jsonify(message=error), 400
    try:
        created = Model(props=data)
    except MissingRequiredAttributeError as err:
        return jsonify(message=err.args[0]), 400
    except (InvalidAttributeError, MismatchROError, ValueError) as err:
        return jsonify(message=err.args[0]), 400
    if result == "object":
        return created
    DB.session.add(created)
    try:
        DB.session.commit()
    except IntegrityError as err:
        DB.session.rollback()
        return jsonify(message="Object violates database constraints", error=err.orig.args[1]), 400
    ID = created.ID
    return jsonify(Model.optimized_query(2).filter(Model.ID == ID).first().fulldesc()), 201


def defaultDelete(Model, ID, name, filters=()):
    """Delete instance with specified ID from the model.

    If no object with the ID exists, a HTTP 404 error is returned.
    If deletion of the object would violate database constraints, a HTTP 400 error is returned.

    Parameters
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to delete from.
    ID : int
        ID of the object to delete.
    name : str
        Object name to use in messages.

    Returns
    -------
    Response
        Flask response containing the new object data or an error message.
    """
    obj = Model.query.filter(Model.ID == ID, *filters).first()
    if obj is None:
        return jsonify(message=name+" not found"), 404
    try:
        DB.session.delete(obj)
        DB.session.commit()
    except IntegrityError as err:
        return jsonify(message="Deletion violates database constraints.", error=err.args[0]), 400
    return jsonify(message="{} #{} deleted.".format(name, ID))


def defaultBatchDelete(Model, filters=()):
    """Delete a list of instances.

    If an ID is not found, it is ignored.
    If deletion of the object would violate database constraints, a HTTP 400 error is returned.

    Parameters
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to delete from.

    Returns
    -------
    Response
        Flask response containing the new object data or an error message.
    """
    if "ID" not in request.args:
        return jsonify(message="Missing ID list"), 400
    IDs = request.args["ID"].split(",")
    objs = Model.query.filter(Model.ID.in_(IDs), *filters).all()
    IDs = [obj.ID for obj in objs]
    try:
        for obj in objs:
            DB.session.delete(obj)
        DB.session.commit()
    except IntegrityError as err:
        return jsonify(message="Object deletion would violate database constraints", error=err.args[0]), 400
    return jsonify(message="Delete successful.", deleted=IDs)


def defaultListHandler(Model, filters=(), order=None, result="response", automatch=True, autofilter=True, autosort=True,
                       include_count="count", query=None):
    """Handle operations on lists.

    Handles list (GET), create (POST) and batch delete (DELETE) requests for the given model.
    Automatically delegates to appropriate default function according to request method.

    Parameters
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to perform the query on
    filters : iterable, optional
        A list of filter expressions to apply to the query. Only applicable for GET requests. The default is None.
    order : list or Column, optional
        Column(s) to use in an order_by expression. Only applicable for GET requests. The default is None.
    result : str, optional
        Return type. See defaultListQuery for more detail. Default is "response"
    automatch : bool, optional
        Whether to apply automatching. See defaultListQuery for more detail.
    autofilter : boolean, optional
        Whether to apply autofiltering. See DataModel.autofilter for more information. Default is True.
    autosort: boolean, optional
        Whether to apply autosorting. See DataModel.autosort for more information. Default is True.
    include_count: str, optional
        Forwarded to defaultListQuery. Default is "count"
    query: BaseQuery, optional
        Base query forwarded to defaultListQuery. Default is None.
    Returns
    -------
    Response
        Flask response containing data or error message.
    """
    if request.method == "GET":
        return defaultListQuery(Model, filters, order, result, automatch, autofilter, autosort, include_count, query)
    elif request.method == "POST":
        return defaultCreate(Model, result)
    elif request.method == "DELETE":
        return defaultBatchDelete(Model, filters)


def defaultObjectHandler(Model, ID, name, filters=()):
    """Handle operations on objects.

    Handles detail (GET), update (PATCH) or delete (DELETE) requests.
    Automatically delegates to appropriate default function according to request method.

    Parameters
    ----------
    Model : SQLAlchemy model with DataModel extension
        Model to perform query on.
    ID : int
        ID of the object to operate on.
    name : str
        Object name to use in messages.

    Returns
    -------
    Response
        Flask response containing data or error message.
    """
    if request.method == "GET":
        return defaultDetailQuery(Model, ID, name, filters)
    elif request.method == "PATCH":
        return defaultPatch(Model, ID, name, None, filters)
    elif request.method == "DELETE":
        return defaultDelete(Model, ID, name, filters)


def userQuery(domainID=None):
    """Query users.

    Parameters
    ----------
    domainID : int, optional
        Filter by domain ID. The default is None.

    Returns
    -------
    Response
        JSON response containing user data
    """
    from sqlalchemy import or_, String
    from sqlalchemy.orm import aliased
    from orm.users import Users, UserProperties
    from tools.constants import PropTags
    from tools.DataModel import DataModel
    from tools.misc import createMapping

    Users._init()
    verbosity = int(request.args.get("level", 1))
    filters = (Users.domainID == domainID,) if domainID is not None else ()
    filters += (Users.ID > 0,)
    query, limit, offset, _ = defaultListHandler(Users, filters=filters, result="query", include_count=None, automatch=False)
    sorts = request.args.getlist("sort")
    for s in sorts:
        sprop, sorder = s.split(",", 1) if "," in s else (s, "asc")
        if hasattr(PropTags, sprop.upper()):
            up = aliased(UserProperties)
            query = query.join(up, (up.userID == Users.ID) & (up.tag == getattr(PropTags, sprop.upper())))\
                         .order_by(up._propvalstr.desc() if sorder == "desc" else up._propvalstr.asc())

    if "match" in request.args:
        expr = request.args["match"]
        fields = set(request.args["matchFields"].split(",")) if "matchFields" in request.args else None
        isUnicode = any(ord(c) > 127 for c in expr)
        matchexpr = tuple("%"+substr+"%" for substr in expr.split())
        matchables = Users._meta.matchables if fields is None else (m for m in Users._meta.matchables if m.alias in fields)
        targets = []
        for prop in matchables:
            column, query = prop.resolve(Users, query)
            if not (isUnicode and isinstance(column.type, String) and column.type.charset == "ascii"):
                targets.append((prop, column))
        if "matchProps" in request.args:
            metaProp = DataModel.Prop(None)
            for prop in request.args["matchProps"].split(","):
                if hasattr(PropTags, prop.upper()):
                    up = aliased(UserProperties)
                    query = query.outerjoin(up, (up.userID == Users.ID) & (up.tag == getattr(PropTags, prop.upper())))
                    targets.append((metaProp, up._propvalstr))
        filters = [column.ilike(match) for match in matchexpr for prop, column in targets if prop.match == "default"] +\
                  [column == prop.tf(expr) for prop, column in targets if prop.match == "exact" and prop.tf(expr) is not None]
        query = query.filter(or_(filter for filter in filters) if filters else False).reset_joinpoint()

    count = query.count()
    data = [user.todict(verbosity) for user in query.limit(limit).offset(offset).all()]
    if verbosity < 2 and "properties" in request.args:
        tags = [getattr(PropTags, prop.upper(), None) for prop in request.args["properties"].split(",")]
        for user in data:
            user["properties"] = {}
        usermap = createMapping(data, lambda x: x["ID"])
        properties = UserProperties.query.filter(UserProperties.userID.in_(usermap.keys()), UserProperties.tag.in_(tags)).all()
        for prop in properties:
            usermap[prop.userID]["properties"][prop.name] = prop.val
    return jsonify(count=count, data=data)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import ServiceHub

import subprocess
import time

from datetime import datetime

def handleSystemdExceptions(service, error):
    if isinstance(error, FileNotFoundError):
        return  # Invalid argument, pass exception on to the caller...
    if isinstance(error, OSError):  # Everything else is probably due to misconfiguration
        return ServiceHub.ERROR


@ServiceHub.register("systemd", handleSystemdExceptions)
class Systemd:
    valmap = {"ActiveState": "state",
              "SubState": "substate",
              "UnitFileState": "autostart",
              "Description": "description",
              "ActiveEnterTimestampMonotonic": "sa",
              "InactiveEnterTimestampMonotonic": "si",
              "Names": "unit"}

    def __init__(self, system=None):
        from tools.config import Config
        self.system = system if system is not None else not Config["options"].get("systemdUser", False)

    @property
    def __mode(self):
        return "--system" if self.system else "--user"

    def getServices(self, *services):
        args = ("systemctl", "-q", self.__mode, "show",
                "--property="+",".join(self.valmap), *services)
        result = subprocess.run(args, stdout=subprocess.PIPE, universal_newlines=True)
        split = [[line.split("=", 1) for line in block.split("\n") if "=" in line] for block in result.stdout.split("\n\n")]
        units = [{self.valmap[key]: value for key, value in block if key in self.valmap} for block in split]
        for unit in units:
            unit["unit"] = unit["unit"].split(" ")[0]
            since = unit["sa"] if unit["state"] == "active" else unit["si"]
            try:
                since = time.clock_gettime(time.CLOCK_REALTIME)-time.clock_gettime(time.CLOCK_MONOTONIC)+int(since)/1000000
                since = datetime.fromtimestamp(int(since)).strftime("%Y-%m-%d %H:%M:%S") if since != 0 else None
            except Exception:
                since = None
            unit["since"] = since
            unit.pop("sa", None), unit.pop("si", None)
        return {unit["unit"]: unit for unit in units if "unit" in unit}

    def run(self, command, *targets):
        result = subprocess.run(("systemctl", "-q", self.__mode, command, *targets),
                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
        return result.returncode, result.stdout

    def startService(self, *services):
        return self.run("start", *services)

    def stopService(self, *services):
        return self.run("stop", *services)

    def restartService(self, *services):
        return self.run("restart", *services)

    def reloadService(self, *services):
        return self.run("reload", *services)

    def tryReloadRestartService(self, *services):
        return self.run("try-reload-or-restart", *services)

    def enableService(self, *services):
        return self.run("enable", *services)

    def disableService(self, *services):
        return self.run("disable", *services)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from cryptography import x509
from cryptography.hazmat.backends import default_backend

from datetime import datetime, MAXYEAR

from .config import Config
from .misc import GenericObject, createMapping

import logging
logger = logging.getLogger("license")


class CertificateError(Exception):
    pass


class GrommunioLicense(GenericObject):
    @staticmethod
    def validate(cert):
        if cert is not None and not cert.not_valid_before <= datetime.now() <= cert.not_valid_after:
            raise CertificateError("Certificate expired")

    @property
    def error(self):
        try:
            self.validate(self.cert)
        except CertificateError as err:
            return err.args[0]


def _defaultLicense():
    return GrommunioLicense(cert=None,
                            file=None,
                            users=5,
                            product="Community",
                            notBefore=datetime(1000, 1, 1),
                            notAfter=datetime(MAXYEAR, 12, 31, 23, 59, 59))


def _processCertificate(data):
    try:
        cert = x509.load_pem_x509_certificate(data, default_backend())
        GrommunioLicense.validate(cert)
        exts = createMapping(cert.extensions, lambda x: x.oid.dotted_string, lambda x: x.value.value)
        lic = GrommunioLicense(cert=cert, file=data)
        lic.users = int(exts.get("1.3.6.1.4.1.56504.1.1"))
        lic.product = exts.get("1.3.6.1.4.1.56504.1.2").decode("utf-8") if "1.3.6.1.4.1.56504.1.2" in exts else None
        lic.notBefore = cert.not_valid_before
        lic.notAfter = cert.not_valid_after
        return True, lic
    except ValueError:
        return False, "Bad certificate"
    except CertificateError as err:
        return False, err.args[0]
    except BaseException as err:
        logger.error(str(err))
        return False, "Unknown error"


def loadCertificate():
    try:
        with open(Config["options"]["licenseFile"], "rb") as file:
            data = file.read()
        success, val = _processCertificate(data)
        if not success:
            logger.error("Failed to load license: "+val)
        else:
            return val
    except KeyError:
        logger.warn("Could not load license: location not configured")
    except FileNotFoundError as err:
        logger.warn("Could not load license: "+err.args[1])


_license = loadCertificate() or _defaultLicense()


def updateCertificate(data):
    success, val = _processCertificate(data)
    if not success:
        return val
    try:
        with open(Config["options"]["licenseFile"], "wb") as file:
            file.write(data)
        global _license
        _license = val
    except KeyError:
        return "Could not load license: location not configured"
    except FileNotFoundError as err:
        return "Could not load license: "
    except PermissionError as err:
        return err.args[1]


def getLicense():
    global _license
    if _license.error:
        _license = _defaultLicense()
    return _license



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import Cli, InvalidUseError
from argparse import ArgumentParser

_typemap = {"normal": 0, "domain": 2}
_typenames = {value: key for key, value in _typemap.items()}
_privmap = {"all": 0, "internal": 1, "domain": 2, "specific": 3, "outgoing": 4}
_privnames = {value: ky for ky, value in _privmap.items()}


def _mkMlistQuery(args):
    from orm.mlists import MLists
    mspec = args.mlistspec
    query = MLists.query
    if mspec is not None:
        query = query.filter(MLists.ID == mspec if mspec.isdigit() else MLists.listname.ilike(mspec+"%"))
    if "filter" in args and args.filter is not None:
        query = MLists.autofilter(query, {f.split("=", 1)[0]: f.split("=", 1)[1] for f in args.filter if "=" in f})
    if "sort" in args and args.sort is not None:
        query = MLists.autosort(query, args.sort)
    return query


def cliMlistCreate(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    from orm.mlists import MLists
    data = dict(listname=args.name,
                listType=_typemap.get(args.type or "normal", 0),
                listPrivilege=_privmap.get(args.privilege or "all", 0),
                associations=args.recipient or [],
                specifieds=args.sender or [])
    error = MLists.checkCreateParams(data)
    if error is not None:
        cli.print(cli.col("Cannot create mlist: "+error, "red"))
        return 3
    try:
        mlist = MLists(data)
        DB.session.add(mlist)
        DB.session.commit()
        cli.print(cli.col("{} ({}):".format(mlist.listname, mlist.ID), attrs=["bold"]))
        _dumpMlist(cli, mlist, 2)
    except ValueError as err:
        cli.print(cli.col("Cannot create mlist: "+err.args[0], "red"))
        DB.session.rollback()


def _dumpMlist(cli, mlist, indent=0):
    cli.print("{}ID: {}".format(" "*indent, mlist.ID))
    cli.print("{}listname: {}".format(" "*indent, mlist.listname))
    col, attr = (None, ["dark"]) if mlist.listType in _typenames else ("red", None)
    cli.print("{}listType: {} ({})".format(" "*indent, mlist.listType,
                                           cli.col(_typenames.get(mlist.listType, "unknown"), col, attrs=attr)))
    col, attr = (None, ["dark"]) if mlist.listPrivilege in _privnames else ("red", None)
    cli.print("{}listPrivilege: {} ({})".format(" "*indent, mlist.listPrivilege,
                                                cli.col(_privnames.get(mlist.listPrivilege, "unknown"), col, attrs=attr)))
    cli.print("{}userID: {}".format(" "*indent, "(not found)" if mlist.user is None else mlist.user.ID))
    if mlist.listType == 0:
        cli.print(" "*indent+"recipients: "+(cli.col("(none)", attrs=["dark"]) if len(mlist.associations) == 0 else ""))
        for assoc in mlist.associations:
            cli.print(" "*indent+"  "+assoc.username)
    if mlist.listPrivilege == 3:
        cli.print(" "*indent+"senders: "+(cli.col("(none)", attrs=["dark"]) if len(mlist.specifieds) == 0 else ""))
        for spec in mlist.specifieds:
            cli.print(" "*indent+"  "+spec.username)


def cliMlistShow(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    mlists = _mkMlistQuery(args).all()
    if len(mlists) == 0:
        cli.print(cli.col("No mailing lists found.", "yellow"))
        return 1
    for mlist in mlists:
        cli.print(cli.col("{} ({}):".format(mlist.listname, mlist.ID), attrs=["bold"]))
        _dumpMlist(cli, mlist, 2)
    DB.session.rollback()


def cliMlistList(args):
    cli = args._cli
    cli.require("DB")
    mlists = _mkMlistQuery(args).all()
    if len(mlists) == 0:
        cli.print(cli.col("No mailing lists found.", "yellow"))
        return 1
    for mlist in mlists:
        cli.print("{}:\t{}".format(mlist.ID, mlist.listname))


def cliMlistDelete(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    mlists = _mkMlistQuery(args).all()
    if len(mlists) == 0:
        cli.print(cli.col("No mailing lists found.", "yellow"))
        return 1
    if len(mlists) > 1:
        cli.print(cli.col("'{}' is ambiguous".format(args.mlistspec), "yellow"))
        return 2
    mlist = mlists[0]
    if not args.yes:
        if cli.confirm("Delete mailing list '{}' ({})? [y/N]: ".format(cli.col(mlist.listname, attrs=["bold"]), mlist.ID))\
                != Cli.SUCCESS:
            return 3
    else:
        cli.print("Deleting mlist '{}' ({})".format(mlist.listname, mlist.ID))
    DB.session.delete(mlist)
    DB.session.commit()
    cli.print("Done.")


def cliMlistAr(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    mlists = _mkMlistQuery(args).all()
    if len(mlists) == 0:
        cli.print(cli.col("No mailing lists found.", "yellow"))
        return 1
    if len(mlists) > 1:
        cli.print("'{}' is ambiguous. Candidates are:".format(args.mlistspec))
        for mlist in mlists:
            cli.print("  {}:\t{}".format(mlist.ID, mlist.listname))
        return 2
    mlist = mlists[0]
    attr = mlist.specifieds if args.attribute == "sender" else mlist.associations
    if args.__command == "remove":
        for element in attr:
            if element.username == args.entry:
                attr.remove(element)
                break
        else:
            cli.print(cli.col("Entry '{}' not found".format(args.entry)))
            return 0
        cli.print("Removed '{}' as {} from {}.".format(args.entry, args.attribute, mlist.listname))
    elif args.__command == "add":
        if sum((1 for element in attr if element.username == args.entry)):
            cli.print(cli.col("'{}' already exists as {} {}".format(args.entry, args.attribute, mlist.listname), "yellow"))
            return 0
        try:
            from orm.mlists import Associations, Specifieds
            attr.append((Specifieds if args.attribute == "sender" else Associations)(args.entry))
            cli.print("Added '{}' as {} to {}".format(args.entry, args.attribute, mlist.listname))
        except ValueError as err:
            cli.print(cli.col(err.args[0], "red"))
            DB.session.rollback()
            return 1
    DB.session.commit()


def cliMlistModify(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    mlists = _mkMlistQuery(args).all()
    if len(mlists) == 0:
        cli.print(cli.col("No mailing lists found.", "yellow"))
        return 1
    if len(mlists) > 1:
        cli.print("'{}' is ambiguous. Candidates are:".format(args.mlistspec))
        for mlist in mlists:
            cli.print("  {}:\t{}".format(mlist.ID, mlist.listname))
        return 2
    mlist = mlists[0]
    attrs = dict()
    if args.privilege is not None:
        attrs["listPrivilege"] = _privmap[args.privilege]
    if args.type is not None:
        attrs["listType"] = _typemap[args.type]
    try:
        mlist.fromdict(attrs)
        DB.session.commit()
    except ValueError as err:
        cli.print(cli.col(err.args[0], "red"))
        DB.session.rollback()
        return 1
    _dumpMlist(cli, mlist)


def _cliListspecCompleter(prefix, **kwargs):
    from orm.mlists import MLists
    return (mlist.listname for mlist in MLists.query.filter(MLists.listname.ilike(prefix+"%"))
                                                    .with_entities(MLists.listname).all())


def _setupCliMlist(subp: ArgumentParser):
    def arArgs(ar, command):
        ar.set_defaults(_handle=cliMlistAr, __command=command)
        ar.add_argument("mlistspec", help="Mlist ID or name").completer = _cliListspecCompleter
        ar.add_argument("attribute", choices=("recipient", "sender"), help="Attribute to modify")
        ar.add_argument("entry", help="Which entry to "+command)

    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    add = sub.add_parser("add", help="Add recipient or sender")
    arArgs(add, "add")
    create = sub.add_parser("create", help="Create new list")
    create.set_defaults(_handle=cliMlistCreate)
    create.add_argument("name", help="Name of the mailing list")
    create.add_argument("-p", "--privilege", choices=_privmap, default="all", help="List privilege type")
    create.add_argument("-r", "--recipient", action="append", help="Users to associate with normal lists")
    create.add_argument("-s", "--sender", action="append",
                        help="Users allowed to send to the list ('specific' privilege only)")
    create.add_argument("-t", "--type", choices=_typemap, help="Mailing list type")
    delete = sub.add_parser("delete", help="Delete list")
    delete.set_defaults(_handle=cliMlistDelete)
    delete.add_argument("mlistspec", help="List ID or name prefix").completer = _cliListspecCompleter
    delete.add_argument("-y", "--yes", action="store_true", help="Do not ask for confirmation")
    list = sub.add_parser("list", help="List existing lists")
    list.set_defaults(_handle=cliMlistList)
    list.add_argument("mlistspec", nargs="?", help="List ID or name prefix")
    list.add_argument("-f", "--filter", nargs="*", help="Filter by attribute, e.g. -f ID=42")
    list.add_argument("-s", "--sort", nargs="*", help="Sort by attribute, e.g. -s listname,desc")
    modify = sub.add_parser("modify", help="Modify list")
    modify.set_defaults(_handle=cliMlistModify)
    modify.add_argument("mlistspec", help="List ID or name prefix").completer = _cliListspecCompleter
    modify.add_argument("-p", "--privilege", choices=_privmap, help="List privilege type")
    modify.add_argument("-t", "--type", choices=_typemap, help="Mailing list type")
    remove = sub.add_parser("remove", help="Remove recipient or sender")
    arArgs(remove, "remove")
    show = sub.add_parser("show", help="Show detailed information about list")
    show.set_defaults(_handle=cliMlistShow)
    show.add_argument("mlistspec", help="Mlist ID or name").completer = _cliListspecCompleter
    show.add_argument("-f", "--filter", nargs="*", help="Filter by attribute, e.g. -f ID=42")
    show.add_argument("-s", "--sort", nargs="*", help="Sort by attribute, e.g. -s listname,desc")


@Cli.command("mlist", _setupCliMlist, help="Mailing/distribution list management")
def cliMlistStub():
    raise InvalidUseError()
